import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import javax.microedition.io.Connector;

import com.nttdocomo.lang.UnsupportedOperationException;
import com.nttdocomo.opt.ui.Graphics2;
import com.nttdocomo.ui.Font;
import com.nttdocomo.ui.IApplication;
import com.nttdocomo.ui.Canvas;
import com.nttdocomo.ui.Graphics;
import com.nttdocomo.ui.Frame;
import com.nttdocomo.ui.Display;
import com.nttdocomo.ui.Image;
import com.nttdocomo.util.Phone;

/**
 * @author kenken
 *
 */
public class Main extends IApplication {

	public void start() {
		MainCanvas mainCanvas = new MainCanvas();
		Display.setCurrent((Frame) (mainCanvas));
		mainCanvas.run();
		terminate();
	}
}

/**
 * @author kenken
 *
 */
class MainCanvas extends Canvas {

	private final NESFile nesFile;

	private final NES nes;

	private final PAD pad;

	private final static int FPS = 60;

	private final static int INTERVAL = 1000 / FPS;
	
	private long lastTime = 0;

	private PPU ppu;
	
	private PSGi psgi;

	private NESApu nesApu;

	protected boolean isAuthError = false;
	
	private int messageUpdateCount = 0;
	
	private long appTime = 0;
	
	private int flameCount = 0;
	
	private boolean doStateSave = false;
		
	private boolean doStateLoad = false;
	
	private boolean isMenuCanceled = false;
	
	/**
	 * 
	 */
	MainCanvas() {
		
		if (this.getWidth() > this.getHeight()) {
			Config.setupKeysHorizontial();
		}
		
		// Config
		Config.load();
		
		// NESFile
		this.nesFile = new NESFile();
		NES.LoadRom(nesFile);
		
		// PAD
		this.pad = new PAD();
		
		// APU
		this.psgi = new PSGi();
		this.nesApu = new NESApu(psgi);
		setSound();
		
		// PPU
		this.ppu = new PPU(nesFile, this.getWidth(), this.getHeight());
		
		// NES
		this.nes = new NES(ppu, pad, nesFile, nesApu);
		
		if (!Phone.getProperty(Phone.USER_ID).equalsIgnoreCase(nes.getUSER_ID())) {
			isAuthError = true;
		}		
		
		setBackground(Graphics.getColorOfName(Graphics.BLACK));

		// x\
		try {
			this.setSoftLabelVisible(false);
		} catch (Exception e) { }
				
		//setLabel();
	}

	public void paint(Graphics g) {
	}

	public void processEvent(int type, int param) {
	}
	
//	private void setLabel()
//	{
//		setSoftLabel(SOFT_KEY_1, "MENU");
//	}

	public void run() {
		Graphics g = getGraphics();
		setClipRect(g);	
		lastTime = System.currentTimeMillis();
		int vSyncWaitCount = 1;
	
		while (true) {
			
			long startAppTime = System.currentTimeMillis();
			
			// L[
			int keyState = getKeypadState();
			if (!isAuthError) {
				pad.ReadKeys(keyState);
			}
			
			if ((keyState & Config.keyMenu) != 0) {
				if (!isMenuCanceled) {
					dispMenu(g);
				}
			} else if ((keyState & Config.keyStateSave) != 0) {								
				doStateSave = stateSave();
			} else if ((keyState & Config.keyStateLoad) != 0) {
				doStateLoad = stateLoad();
			} else if ((keyState & Config.keyMenu) == 0) {
				isMenuCanceled = false;
			}
									
			// CPU,PPU̓ԂXV
			nes.UpdateOneFlame();
			
			// OtBbNX`
			g.lock();
			DrawOneFrame(g);
			drawMessage(g);
			drawAuthMessage(g);
			
			// APȔ
			nesApu.doFrame();

			if (Config.sleepEnable) {
				if (Config.vSyncWaitEnable && g instanceof Graphics2) {
					Graphics2 g2 = (Graphics2)g;
					if (vSyncWaitCount <= 1) {
						vSyncWaitCount = g2.syncUnlock(1);
					} else {
						vSyncWaitCount--;
						g.unlock(true);						
						sleep();
					}
				} else {
					g.unlock(true);
					sleep();
				}
			} else {
				g.unlock(true);
			}
			
			appTime += (System.currentTimeMillis() - startAppTime);
		}
	}

	private void sleep() {
		long wait = INTERVAL - (System.currentTimeMillis() - lastTime);
		while (wait > 1) {
			try {
				Thread.sleep(wait / 2);
			} catch (Exception e) {
			}
			wait = INTERVAL - (System.currentTimeMillis() - lastTime);
		}
		lastTime = System.currentTimeMillis();
	}
	
	public void DrawOneFrame(Graphics g)
	{		
		if (flameCount % (Config.frameSkip + 1) == 0) {
			//g.clearRect(0, 0, getWidth(), getHeight());
			ppu.Render(g);
		}				
		flameCount++;
		
		if ((flameCount & 511) == 0) {
			ppu.isDirtyAll = true;
		}
	}
	
	private void setClipRect(Graphics g) {
		g.setClip(ppu.SCREEN_OFFSET_X, ppu.SCREEN_OFFSET_Y + 8,
				this.getWidth() - ppu.SCREEN_OFFSET_X, this.getHeight() - (ppu.SCREEN_OFFSET_Y + 16));
	}
	
	private boolean existsSyncUnlock(Graphics g) {
		try {			
			if (g instanceof Graphics2) {
				Graphics2 g2 = (Graphics2)g;
				g2.lock();
				g2.syncUnlock(1);
				return true;
			}
		} catch (UnsupportedOperationException e) {
			e.printStackTrace();
		}
		return false;
	}

	
	private int nessageCount = 0;
	private String[] message;
	
	private void drawMessage(Graphics g)
	{
		if (Config.dispFPSEnable) {
			if (messageUpdateCount > 60*5) {
				setSoftLabel(SOFT_KEY_2, "FPS" + (1000 * messageUpdateCount / appTime));
				messageUpdateCount = 0;
				appTime = 0;						
			}
			messageUpdateCount++;
		}
		
		if (doStateSave) {
			message = new String[] {
				"",
				"Z[tJ܂", 
				"", 		
			};
			nessageCount = 30;
			doStateSave = false;
		}
		
		if (doStateLoad) {
			message = new String[] {
				"",
				"[gJ܂", 
				"", 		
			};
			nessageCount = 30;
			doStateLoad = false;
		}
		
		if (nessageCount > 0) {
			drawMessages(g, message, 16, 16);
			nessageCount--;
		}
	}
	
	private void drawAuthMessage(Graphics g)
	{
		if (isAuthError) {
			final int FONT_PIXEL = 12;
			g.setColor(Graphics.getColorOfName(Graphics.BLUE));
			g.fillRect(FONT_PIXEL * 3,
					this.getHeight() / 2 - FONT_PIXEL,
					getWidth() - FONT_PIXEL * 6,
					FONT_PIXEL * 5);
			
			g.setColor(Graphics.getColorOfName(Graphics.WHITE));
			
			int msgY = this.getHeight() / 2; 
			
			String msg = "FOMAJ[hԍ̔F؃G[łB";
			g.drawString(msg, 
					this.getWidth() / 2 - FONT_PIXEL * msg.length() / 2,
					msgY);

			msg = "ȂFOMAJ[hԍ́A";
			msgY += 16;
			g.drawString(msg, 
					this.getWidth() / 2 - FONT_PIXEL * msg.length() / 2,
					msgY);
			
			msg = Phone.getProperty(Phone.USER_ID);
			msgY += 16;
			g.drawString(msg,
					(this.getWidth() - 16 * msg.length() / 2) / 2,
					msgY);

			msg = "łB";
			msgY += 16;
			g.drawString(msg,
					(this.getWidth() - 16 * msg.length() / 2) / 2,
					msgY);
			
			if (flameCount > FPS * 60) {
				IApplication.getCurrentApp().terminate();
			}
		}
	}
	
	public static void drawString2(Graphics g, String str, int x, int y) {		
		g.clearRect(x, y, 16 * str.length(), 16);

		// 16dotTCỸtHgw
		Font font = Font.getFont(Font.FACE_SYSTEM | Font.STYLE_PLAIN | Font.SIZE_SMALL);
		g.setFont(font);
		g.setColor(Graphics.getColorOfName(Graphics.WHITE));
		g.drawString(str, x, y + font.getAscent());
	}
	
	public static void drawString2_(Graphics g, String str, int x, int y) {
		final int FONT_DOTS = 16;
		int whiteColor = Graphics.getColorOfName(Graphics.WHITE);
		int blackColor = Graphics.getColorOfName(Graphics.BLACK);
		
		// `obt@pC[W쐬
		Image srcImage = Image.createImage(FONT_DOTS, FONT_DOTS);
		Graphics gSrc = srcImage.getGraphics();
		
		Image dstImage = Image.createImage(FONT_DOTS / 2, FONT_DOTS / 2);
		Graphics gDst = dstImage.getGraphics();
		
		// 16dotTCỸtHgw
		Font font = Font.getFont(Font.FACE_SYSTEM | Font.STYLE_PLAIN | Font.SIZE_SMALL);
		gSrc.setFont(font);
		
		for (int i = 0; i < str.length(); i++) {
			// `obt@ŃNA
			gSrc.setColor(blackColor);
			gSrc.fillRect(0, 0, FONT_DOTS, FONT_DOTS);
			
			// łP`
			gSrc.setColor(whiteColor);
			gSrc.drawString(str.substring(i, i + 1), 0, 16);
			
			// sNZf[^i[zp
			int[] srcPixels = new int[FONT_DOTS * FONT_DOTS];
			int[] dstPixels = new int[(FONT_DOTS / 2) * (FONT_DOTS / 2)];
			gSrc.getPixels(0, 0, FONT_DOTS, FONT_DOTS, srcPixels, 0);
			
			// dstNA
			for (int j = 0; j < dstPixels.length; j++) {
				dstPixels[j] = blackColor;
			}
			
			// 16x168x8ɏk
			for (int sy = 0; sy < FONT_DOTS; sy++) {
				for (int sx  = 0; sx < FONT_DOTS; sx++) {
					if (srcPixels[(sy << 4) + sx] == whiteColor) {
						dstPixels[((sy >> 1) << 3) + (sx >> 1)] = whiteColor;
					}
				}
			}
			
			gDst.setColor(blackColor);
			gDst.fillRect(0, 0, dstImage.getWidth(), dstImage.getHeight());
			gDst.setPixels(0, 0, 8, 8, dstPixels, 0);
			g.setFlipMode(Graphics.FLIP_NONE);
			g.drawScaledImage(dstImage,
					x + (i << 4), y,
					16, 16,
					0, 0,
					8, 8);
		}
	}
	
	private void setSound() {
		psgi.setFrameBufCount(Config.APUBuffer);
		psgi.setMasterVolume(Config.APUVolume);
		psgi.setSoundEnable(Config.APUEnable);
		psgi.setSoundStereo(Config.APUStereoEnable);
		psgi.setChEnable(0, true);
		psgi.setChEnable(1, true);
		psgi.setChEnable(2, true);
		psgi.setChEnable(3, true);

		switch (Config.APUMode) {
		case 0:
			psgi.setMode(PSGi.MODE_PRESET);
			psgi.setToneBank(0, 80);
			psgi.setToneBank(1, 80);
			psgi.setToneBank(2, 80);
			psgi.setToneBank(3, 80);
			psgi.setToneBank(4, 127);
			break;
		case 1:
			psgi.setMode(PSGi.MODE_PRESET);
			psgi.setToneBank(0, 81);
			psgi.setToneBank(1, 81);
			psgi.setToneBank(2, 81);
			psgi.setToneBank(3, 81);
			psgi.setToneBank(4, 127);
			break;
		case 2:
			psgi.setMode(PSGi.MODE_YAMAHA);
			psgi.setToneBank(0, 0);
			psgi.setToneBank(1, 1);
			psgi.setToneBank(2, 2);
			psgi.setToneBank(3, 3);
			psgi.setToneBank(4, 4);
			break;
		case 3:
			psgi.setMode(PSGi.MODE_ROHM);
			psgi.setToneBank(0, 0);
			psgi.setToneBank(1, 1);
			psgi.setToneBank(2, 2);
			psgi.setToneBank(3, 3);
			psgi.setToneBank(4, 4);
			break;
		}
	}
	
	public boolean stateLoad(){
		DataInputStream input = null;
		try {
			input = Connector.openDataInputStream("scratchpad:///1;pos=0");
			if (!(input.readChar() == 'S' && input.readChar() == 'S')) {
				return false;
			}
			lastTime = input.readLong();
			flameCount = input.readInt();
			nes.stateLoad(input);
			ppu.stateLoad(input);
			pad.stateLoad(input);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		} finally {
			try {
				input.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}		
		return true;
	}
	
	public boolean stateSave(){
		DataOutputStream output = null;
		try {
			output = Connector.openDataOutputStream("scratchpad:///1;pos=0");
			output.writeChars("SS");
			output.writeLong(lastTime);
			output.writeInt(flameCount);
			nes.stateSave(output);
			ppu.stateSave(output);
			pad.stateSave(output);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		} finally {
			try {
				output.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}
		return true;
	}
	
	private void dispMenu(Graphics g)
	{
		g.clearClip();
		
		String[] messages = new String[] {
			"",
			"@J߂̂Ă@", 
			"AƂ̂Ă@@@", 
			"BL[̂Ă@@@", 
			"C\tgZbg@@@", 
			"DƁJ@@@@@@", 
			"", 		
		};
		
		int cursorNo = 1;
		try		
		{
			while (!isMenuCanceled) {
				String[] dispMessage = replaceCursor(messages, cursorNo);
				drawMessages(g, dispMessage, 16, 16);
				
				int keyState = getKey();
				if ((keyState & Config.keyOK) != 0) {
					switch (cursorNo) {
						case 1:
							dispScreenSetting(g);
							break;
						case 2:
							dispSoundSetting(g);
							break;
						case 3:
							dispKeySetting(g);
							break;
						case 4:
							if (confirm(g)) {
								nes.reset();
							}
							return;
						case 5: 
							return;
					}
				} else if ((keyState & Config.keyCancel) != 0) {
					return;
				} else if ((keyState & Config.keyMenu) != 0) {
					isMenuCanceled = true;
				} else {
					cursorNo = moveCursor(keyState, cursorNo, 5);
				}
			}
		} finally {
			g.clearRect(0, 0, getWidth(), getHeight());
			setClipRect(g);
			Config.save();
		}
	}
	
	private boolean confirm(Graphics g) {
		int cursorNo = 2;
		while (true) {
			String[] messages = new String[] {
				"",
				"낵āJH", 
				"@͂@@@@@@", 
				"A@@@@@", 
				"", 		
			};
					
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 32, 64);	
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						return true;
					case 2:
						return false;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return false;
			} else if ((keyState & Config.keyMenu) != 0) {
				return false;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 2);
			}
		}
	}
	
	private void dispScreenSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@t[XLbv@@", 
				"AƁJ傤@", 
				"BeorЂ傤J@" + (Config.dispFPSEnable ? "d" : "@") + "", 
				"CƁJ@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						dispFrameSkipSetting(g);
						break;
					case 2:
						dispSleepSetting(g);
						break;
					case 3:
						Config.dispFPSEnable = !Config.dispFPSEnable;
						if (!Config.dispFPSEnable) {
							setSoftLabel(SOFT_KEY_2, null);
						}
						try {
							this.setSoftLabelVisible(Config.dispFPSEnable);
						} catch (Exception e) { }
						break;
					case 4: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 4);
			}
		}
	}
	
	private void dispFrameSkipSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@erP@@@@@@" + getFrameSkipStr(0) + "", 
				"AerQ@@@@@@" + getFrameSkipStr(1) + "", 
				"BerR@@@@@@" + getFrameSkipStr(2) + "", 
				"CerS@@@@@@" + getFrameSkipStr(3) + "", 
				"DerT@@@@@@" + getFrameSkipStr(4) + "", 
				"EerU@@@@@@" + getFrameSkipStr(5) + "", 
				"FerV@@@@@@" + getFrameSkipStr(6) + "", 
				"GerW@@@@@@" + getFrameSkipStr(7) + "", 
				"HƁJ@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
					case 2:
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
						Config.frameSkip = cursorNo - 1;
						break;
					case 9: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 9);
			}
		}
	}
	
	private String getFrameSkipStr(int skip) {
		return (Config.frameSkip == skip) ? "d" : "@";
	}
	
	private void dispSleepSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@X[v@@@@@@" + (Config.sleepEnable ? "d" : "@") + "", 
				"Aurxmb܂@@@" + (Config.vSyncWaitEnable ? "d" : "@") + "", 
				"BƁJ@@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						Config.sleepEnable = !Config.sleepEnable;
						break;
					case 2: 
						Config.vSyncWaitEnable = !Config.vSyncWaitEnable;
						if (Config.vSyncWaitEnable && !existsSyncUnlock(g)) {
							Config.vSyncWaitEnable = false;
							dispError(g);
						}
						break;
					case 3: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 3);
			}
		}
	}
	
	private void dispError(Graphics g) {
		String[] messages = new String[] {
			"",
			"݂̂̂ł", 
			"", 		
		};
				
		drawMessages(g, messages, 16, 64);			
		getKey();
	}
	
	private void dispSoundSetting(Graphics g)
	{
		int cursorNo = 1;

		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@񂯁J񃂁[gJ@@", 
				"AzJ[@@@@@", 
				"BnJbt@@@@@@@", 
				"CƂ䂤@@@@" + (Config.APUEnable ? "d" : "@") + "", 
				"DXeI@@@@@@" + (Config.APUStereoEnable ? "d" : "@") + "", 
				"EƁJ@@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
						dispSoundEngineSetting(g);
						break;
					case 2: 
						dispSoundVolumeSetting(g);
						break;
					case 3:
						dispSoundBufferSetting(g);
						break;
					case 4: 
						Config.APUEnable = !Config.APUEnable;
						setSound();
						break;
					case 5: 
						Config.APUStereoEnable = !Config.APUStereoEnable;
						setSound();
						break;
					case 6:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 6);
			}
		}
	}
	
	private void dispSoundEngineSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@[gJP@@@@" + getSoundModeStr(0) + "", 
				"A[gJQ@@@@" + getSoundModeStr(1) + "", 
				"B[gJR@@@@" + getSoundModeStr(2) + "", 
				"C[gJS@@@@" + getSoundModeStr(3) + "", 
				"DƁJ@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
					case 2: 
					case 3: 
					case 4: 
						Config.APUMode = cursorNo - 1;
						setSound();
						break;
					case 5:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 5);
			}
		}
	}
	
	private String getSoundModeStr(int soundMode) {
		return (Config.APUMode == soundMode) ? "d" : "@";
	}
	
	private void dispSoundVolumeSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@zJ[QO@" + getVolumeStr(20 ) + "", 
				"AzJ[SO@" + getVolumeStr(40 ) + "", 
				"BzJ[UO@" + getVolumeStr(60 ) + "", 
				"CzJ[WO@" + getVolumeStr(80 ) + "", 
				"DzJ[l`w@" + getVolumeStr(100) + "", 
				"EƁJ@@@@@@@", 
				"", 		
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
					case 2: 
					case 3: 
					case 4: 
					case 5: 
						Config.APUVolume = cursorNo * 20;
						setSound();
						break;
					case 6:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 6);
			}
		}
	}
	
	private String getVolumeStr(int vol) {
		return (Config.APUVolume == vol) ? "d" : "@";
	}
	
	private void dispSoundBufferSetting(Graphics g)
	{
		int cursorNo = 1;

		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@nJbt@T@@@@"+ getApuBufferStr( 5) + "", 
				"AnJbt@PO@@@"+ getApuBufferStr(10) + "", 
				"BnJbt@PT@@@"+ getApuBufferStr(15) + "", 
				"CnJbt@QO@@@"+ getApuBufferStr(20) + "", 
				"DnJbt@QT@@@"+ getApuBufferStr(25) + "", 
				"EnJbt@RO@@@"+ getApuBufferStr(30) + "", 
				"FnJbt@RT@@@"+ getApuBufferStr(35) + "", 
				"GnJbt@SO@@@"+ getApuBufferStr(40) + "", 
				"HƁJ@@@@@@@", 
				"", 		
			};
				
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 16);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
					case 2: 
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
						Config.APUBuffer = cursorNo * 5;
						setSound();
						break;
					case 9:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 9);
			}
		}
	}
	
	private String getApuBufferStr(int bufSize) {
		return (Config.APUBuffer == bufSize) ? "d" : "@";
	}
	
	private void dispKeySetting(Graphics g)
	{
		int cursorNo = 1;

		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"",
				"@@@" + getKeySettingString(Config.pad1Up) + "", 
				"A@@" + getKeySettingString(Config.pad1Down) + "", 
				"BЂJ" + getKeySettingString(Config.pad1Left) + "", 
				"C݂J@" + getKeySettingString(Config.pad1Right) + "", 
				"D`@@@" + getKeySettingString(Config.pad1A) + "", 
				"Ea@@@" + getKeySettingString(Config.pad1B) + "", 
				"Frs@@" + getKeySettingString(Config.pad1Start) + "", 
				"Grk@@" + getKeySettingString(Config.pad1Select) + "", 
				"HZ[tJ" + getKeySettingString(Config.keyStateSave) + "", 
				"I[gJ" + getKeySettingString(Config.keyStateLoad) + "", 
				"JƁJ@@@@@@", 
				"", 
			};
			
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, 16, 0);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
					case 2: 
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
					case 9:
					case 10:
						dispKeySetting2(g, cursorNo);
						break;
					case 11:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 11);
			}
		}
	}
			
	private void dispKeySetting2(Graphics g, int keyNo) {
		String[] messages = new String[] {
			"",
			"L[ĂJ", 
			"", 		
		};
				
		drawMessages(g, messages, 16, 64);			
		int keyState = getPadKey();
		switch (keyNo) {
			case 1: Config.pad1Up = keyState; break;
			case 2: Config.pad1Down = keyState; break;
			case 3: Config.pad1Left = keyState; break;
			case 4: Config.pad1Right = keyState; break;
			case 5: Config.pad1A = keyState; break;
			case 6: Config.pad1B = keyState; break;
			case 7: Config.pad1Start = keyState; break;
			case 8: Config.pad1Select = keyState; break;
			case 9: Config.keyStateSave = keyState; break;
			case 10: Config.keyStateLoad = keyState; break;
		}
	}
	
	private String getKeySettingString(int padkey) {
		switch (padkey) {
			case (1 << Display.KEY_0): 			return "@@@@@O";
			case (1 << Display.KEY_1): 			return "@@@@@P";
			case (1 << Display.KEY_2): 			return "@@@@@Q";
			case (1 << Display.KEY_3): 			return "@@@@@R";
			case (1 << Display.KEY_4): 			return "@@@@@S";
			case (1 << Display.KEY_5): 			return "@@@@@T";
			case (1 << Display.KEY_6): 			return "@@@@@U";
			case (1 << Display.KEY_7): 			return "@@@@@V";
			case (1 << Display.KEY_8): 			return "@@@@@W";
			case (1 << Display.KEY_9): 			return "@@@@@X";
			case (1 << Display.KEY_ASTERISK): 	return "@@@@@";
			case (1 << Display.KEY_DOWN): 		return "@@@@@";
			case (1 << Display.KEY_IAPP): 		return "@@@@@";
			case (1 << Display.KEY_LEFT): 		return "@@@@@";
			case (1 << Display.KEY_PAGE_DOWN): 	return "@o`fd";
			case (1 << Display.KEY_PAGE_UP): 	return "@o`fd";
			case (1 << Display.KEY_POUND): 		return "@@@@@";
			case (1 << Display.KEY_RIGHT): 		return "@@@@@";
			case (1 << Display.KEY_SELECT):		return "@@@rdk";
			case (1 << Display.KEY_UP): 		return "@@@@@";
		}
		return "";
	}
	
	private int getKey() {
		final int mask = 
			  Config.keyUp
			| Config.keyDown
			| Config.keyOK
			| Config.keyCancel
			| Config.keyMenu;

		int keyState;
		
		do {
			keyState = getKeypadState();
			sleep();
		} while ((keyState & mask) != 0);
		
		do {
			keyState = getKeypadState();
			sleep();
		} while (bitCount(keyState & mask) != 1);
		
		return keyState;
	}
	
	private int getPadKey() {
		final int mask = 
			(1 << Display.KEY_0)
		  | (1 << Display.KEY_1)
		  | (1 << Display.KEY_2)
		  | (1 << Display.KEY_3)
		  | (1 << Display.KEY_4)
		  | (1 << Display.KEY_5)
		  | (1 << Display.KEY_6)
		  | (1 << Display.KEY_7)
		  | (1 << Display.KEY_8)
		  | (1 << Display.KEY_9)
		  | (1 << Display.KEY_ASTERISK)
		  | (1 << Display.KEY_DOWN)
		  | (1 << Display.KEY_IAPP)
		  | (1 << Display.KEY_LEFT)
		  | (1 << Display.KEY_PAGE_DOWN)
		  | (1 << Display.KEY_PAGE_UP)
		  | (1 << Display.KEY_POUND)
		  | (1 << Display.KEY_RIGHT)
		  | (1 << Display.KEY_SELECT)
		  | (1 << Display.KEY_UP);

		int keyState;
		
		do {
			keyState = getKeypadState();
			sleep();
		} while ((keyState & mask) != 0);
		
		do {
			keyState = getKeypadState();
			sleep();
		} while (bitCount(keyState & mask) != 1);
		
		return keyState;
	}
	
	private int bitCount(int bit) {
		int count = 0;
		for (int i = 0; i < 32; i++) {
			if ((bit & (1 << i)) != 0) {
				count++;
			}
		}
		return count;
	}
		 
	private int moveCursor(int keyState, int cursorNo, int maxNo) {
		
		if ((keyState & Config.keyUp) != 0)
		{
			cursorNo--;
		}
		if ((keyState & Config.keyDown) != 0)
		{
			cursorNo++;
		}
		
		if (cursorNo > maxNo) {
			cursorNo = 1;
		} else if (cursorNo < 1) {
			cursorNo = maxNo;
		}
		
		return cursorNo;
	}
	
	private String[] replaceCursor(String[] messages, int cursorNo) {
		final char[] numChars = new char[] {
			'@','@','A','B','C','D','E','F','G','H','I','J'
		};
			
		String[] ret = new String[messages.length];
		
		for (int i = 0; i < messages.length; i++) {
			ret[i] = messages[i].replace(numChars[cursorNo], '');
			for (int j = 1; j < numChars.length; j++) {
				ret[i] = ret[i].replace(numChars[j], '@');
			}
		}		
		return ret;
	}
	
	private void drawMessages(Graphics g, String[] messages, int x, int y) {
		g.lock();
		for (int i = 0; i < messages.length; i++) {
			drawString2(g, messages[i], x, y += 16);
		}
		g.unlock(true);
	}
}
