import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * NESn[hEFAiCPU,PPU,WRAMjG~[V邽߂̃NXłB
 * @author aoisome
 */
class NES {
	protected int PC;

	private int CLK;

	private int A;

	private int T;

	private int X;

	private int Y;

	private int SP;

	private int C;

	private int I;

	private int D;

	private int B;

	private int R;

	private int V;
		
	protected int DMAclk = 0;
	
	protected final int prgMask;
	
	protected final int[] wram;

	protected final int[] prgRom;

	private final PPU ppu;
	
	private final PAD pad;

	private final NESApu nesApu;		
	
	/**
	 * NESG~[V̂߂̏s܂B
	 * @param pad
	 * @param nesFile
	 * @param nesApu
	 */
	NES(PPU ppu, PAD pad, NESFile nesFile, NESApu nesApu) {		
		// CPU
		this.SP = 0xff;
		this.CLK = 0;

		// MEM
		this.wram = new int[0x0800];
		this.prgRom = nesFile.getPrgRom();
		this.prgMask = nesFile.getPrgSize() - 1;

		// PPU
		this.ppu = ppu;
		
		// PAD
		this.pad = pad;
		
		// APU
		this.nesApu = nesApu;
		this.nesApu.init();

		// RESET
		R = SFLG.R;
		PC = LdMemW(0xFFFC);
	}
	
	public final void reset() {
		// RESET
		R = SFLG.R;
		PC = LdMemW(0xFFFC);
	}

	/**
	 * NES̓Ԃ1t[AXV܂B 
	 */
	public final void UpdateOneFlame() {
		final int CYCLES_PER_LINE = 113;
		
		ppu.ClearPPUWriteBuffer();
		ppu.StorePaletteBack();		

		int scanLine;
		
		for (scanLine = 0; scanLine < 240; scanLine++) {
			ppu.setScanLine(scanLine);
			ppu.CheckSpriteZeroHit();
			SetClk(GetClk() + CYCLES_PER_LINE);
			ExecuteFast();			
		}

		for (scanLine = 240; scanLine <= 261; scanLine++) {
			ppu.CheckSpriteZeroHit();
			
			if (scanLine == 261) {
				ppu.ClearVBlank();
			}

			if (scanLine == 241) {
				ppu.SetVBlank();
				ppu.ClearSpriteZeroHit();

				SetClk(GetClk() + 1);
				ExecuteFast();
				ppu.TrigNMI(this);

				SetClk(GetClk() + CYCLES_PER_LINE - 1);
				ExecuteFast();
				continue;
			}

			SetClk(GetClk() + CYCLES_PER_LINE);
			ExecuteFast();
		}
	}

	public void stateLoad(DataInputStream input) throws IOException {
		PC = input.readInt();
		CLK = input.readInt();
		A = input.readUnsignedByte();
		T = input.readUnsignedByte();
		X = input.readUnsignedByte();
		Y = input.readUnsignedByte();
		SP = input.readUnsignedByte();
		C = input.readUnsignedByte();
		I = input.readUnsignedByte();
		D = input.readUnsignedByte();
		B = input.readUnsignedByte();
		R = input.readUnsignedByte();
		V = input.readUnsignedByte();
		for (int i = 0; i < wram.length; i++) {
			wram[i] = input.readUnsignedByte();
		}
	}
	  
	public void stateSave(DataOutputStream output) throws IOException {
		output.writeInt(PC);
		output.writeInt(CLK);
		output.writeByte(A);
		output.writeByte(T);
		output.writeByte(X);
		output.writeByte(Y);
		output.writeByte(SP);
		output.writeByte(C);
		output.writeByte(I);
		output.writeByte(D);
		output.writeByte(B);
		output.writeByte(R);
		output.writeByte(V);
		for (int i = 0; i < wram.length; i++) {
			output.writeByte(wram[i]);
		}
	}

	/**
	 * ݂̎csNbNԂ܂B
	 * @return
	 */
	public final int GetClk() {
		return CLK;
	}

	/**
	 * csNbNݒ肵܂B
	 * @param clk
	 */
	public final void SetClk(int clk) {
		CLK = clk;
	}

	/**
	 * NMI𔭐܂B
	 */
	public final void NMI() {
		B = 0;
		PushStackW(PC);
		PushStackB(BindFlags());
		I = SFLG.I;
		PC = LdMemW(0xFFFA);
	}

	/**
	 * IRQ𔭐܂B
	 */
	public final void IRQ() {
		if ((I & SFLG.I) != 0) {
			return;
		}

		B = 0;
		PushStackW(PC);
		PushStackB(BindFlags());
		I = SFLG.I;
		PC = LdMemW(0xFFFE);
	}

	/**
	 * tOWX^\intϐ̃rbgqĕԂ܂B
	 * @return
	 */
	private final int BindFlags() {
		return C | ((T <= 0) ? SFLG.Z : 0) | I | D | B | R
				| (((V & 0x80) != 0) ? SFLG.V : 0)
				| (((T & 0x80) != 0) ? SFLG.N : 0);
	}

	/**
	 * w肵intϐ̃rbgAtOWX^\ϐ֓WJ܂B
	 * @param p
	 */
	private final void UnbindFlags(int p) {
		C = p & SFLG.C;
		I = p & SFLG.I;
		D = p & SFLG.D;
		B = p & SFLG.B;
		R = p & SFLG.R;
		V = p << 1;
		T = ((p & 0x82) ^ 2) | ((p & 2) << 30);
	}

	/* MEM */

	/**
	 * w肵AhXCPUԂA1oCgǂݍ܂B
	 * @param address
	 * @return
	 */
	protected final int LdMemB(int address) {
		if (address <= 0x1FFF) {
			// WRAM
			return wram[address & 0x7FF];
		} else if (address > 0x7FFF) {
			return prgRom[address & prgMask];
		} else if (address <= 0x3FFF) {
			// PPU I/O
			if ((address & 0x7) == 2) {
				return ppu.GetStatusRegister();
			} else if ((address & 0x7) == 7) {
				return ppu.GetPPUMemoryData();
			}
			return 0;
		} else if (address <= 0x401F) {
			// APU I/OAPAD
			if ((address & 0x1F) == 0x16) {
				return pad.GetPad1(); // PAD1
			} else if ((address & 0x1F) == 0x17) {
				return pad.GetPad2(); // PAD2
			}
			return nesApu.readReg();
		}
//		else if (address <= 0x5FFF) {
//			// gROM
//			return 0; // dummy
//		} else if (address <= 0x7FFF) {
//			// gRAM
//			return 0; // dummy
//		}
		return 0;
	}

	/**
	 * w肵AhXCPUԂA1oCgǂݍ݂܂B
	 * PRG-ROMǂޏꍇɁÃ\bhgpƁAłB
	 * @param address
	 * @return
	 */
	private final int LdMemB_PRG(int address) {
		if (address > 0x7FFF) {
			return prgRom[address & prgMask];
		} else if (address <= 0x1FFF) {
			// WRAM
			return wram[address & 0x7FF];
		} else if (address <= 0x3FFF) {
			// PPU I/O
			if ((address & 0x7) == 2) {
				return ppu.GetStatusRegister();
			} else if ((address & 0x7) == 7) {
				return ppu.GetPPUMemoryData();
			}
			return 0;
		} else if (address <= 0x401F) {
			// APU I/OAPAD
			if ((address & 0x1F) == 0x16) {
				return pad.GetPad1();			// PAD1
			} else if ((address & 0x1F) == 0x17) {
				return pad.GetPad2(); // PAD2
			}
			return nesApu.readReg();
		}
//		else if (address <= 0x5FFF) {
//			// gROM
//			return 0; // dummy
//		} else if (address <= 0x7FFF) {
//			// gRAM
//			return 0; // dummy
//		}
		return 0;
	}

	/**
	 * w肵AhXCPUԂA2oCgǂݍ݂܂B
	 * @param address
	 * @return
	 */
	private final int LdMemW(int address) {
		if (address > 0x7FFF) {
			return prgRom[address & prgMask]
					| (prgRom[(address + 1) & prgMask] << 8);
		} else if (address <= 0x1FFF) {
			return wram[address & 0x7FF] | (wram[(address + 1) & 0x7FF] << 8);
		} else {
			return LdMemB(address) | (LdMemB(address + 1) << 8);
		}
	}

	/**
	 * w肵AhXCPŨԂA2oCgǂݍ݂܂B
	 * ꉞAPRG-ROMǂނ߂Ɏgp郁\bhłA󂻂ȂĂȂłB
	 * @param address
	 * @return
	 */
	private final int LdPrgRomW(int address) {
		if (address > 0x7FFF) {
			return prgRom[address & prgMask]
					| (prgRom[(address + 1) & prgMask] << 8);
		} else if (address <= 0x1FFF) {
			return wram[address & 0x7FF] | (wram[(address + 1) & 0x7FF] << 8);
		} else {
			return LdMemB(address) | (LdMemB(address + 1) << 8);
		}
	}
	
	/**
	 * w肵AhXCPUԂɑ΂āAw肵l݂܂B
	 * @param address
	 * @param val
	 */
	private final void StMem(int address, int val) {
		if (address < 0x1FFF) {
			// WRAM
			wram[address & 0x7FF] = val;
		} else if (address <= 0x3FFF) {
			// PPU I/O
			switch (address & 0x7) {
			case 0:
				ppu.SetControlRegister1(val);
				break;
			case 1:
				ppu.SetControlRegister2(val);
				break;
			case 2:
				break;
			case 3:
				ppu.SetSpriteMemoryAddress(val);
				break;
			case 4:
				ppu.SetSpriteMemoryData(val);
				break;
			case 5:
				ppu.SetBGScrollOffset(val);
				break;
			case 6:
				ppu.SetPPUMemoryAddress(val);
				break;
			case 7:
				ppu.SetPPUMemoryData(val);
				break;
			}
		} else if (address <= 0x401F) {
			// APU I/OAPAD
			switch (address & 0x1F) {
			case 0x14: // SPRITE DMA
				ppu.DMA(this, val);
				break;
			case 0x16: // PAD1
				pad.SetPad1(val);
				break;
			case 0x17: // PAD2
				break;
			default:
				nesApu.writeReg(address, val);
				break;
			}
		}
		// else if (address <= 0x5FFF)
		// {
		// // gROM
		// }
		// else if (address <= 0x7FFF)
		// {
		// // gRAM
		// }
		// else if (address <= 0xBFFF)
		// {
		// // PRG-ROM
		// }
		// else if (address <= 0xFFFF)
		// {
		// // PRG-ROM
		// }
	}

	/**
	 * w肵AhXWRAMA2oCgǂݍ݂܂B
	 * @param address
	 * @return
	 */
	private final int LdWramW(int address) {
		return wram[address & 0x7FF] | (wram[(address + 1) & 0x7FF] << 8);
	}

	/**
	 * X^bNɎw肵	1oCg̒lvbV܂B
	 * @param val
	 */
	private final void PushStackB(int val) {
		wram[SP + 256] = val;
		SP = (SP - 1) & 0xFF;
	}

	/**
	 * X^bNoCg̒lv܂B
	 * @return
	 */
	private final int PullStackB() {
		return wram[((SP += 1) & 0xFF) + 256];
	}

	/**
	 * X^bNɎw肵2oCg̒lvbV܂B
	 * @param val
	 */
//	private final void PushStackW(int val) {
//		wram[SP + 256] = val >> 8;
//		SP = (SP - 1) & 0xff;
//		wram[SP + 256] = val & 0xFF;
//		SP = (SP - 1) & 0xff;
//	}
	private final void PushStackW(int val) {
		wram[SP + 0x100] = val >> 8;	
		wram[(SP + 0xFF) | 0x100] = val & 0xFF;
		SP = (SP - 2) & 0xFF;
	}

	/**
	 * X^bN2oCg̒lv܂B
	 * @return
	 */
	private final int PullStackW() {
		int ret = (wram[(SP + 2) | 0x100] << 8) | wram[(SP + 1) | 0x100];
		SP = (SP + 2) & 0xFF;
		return ret;
	}


	/**
	 * csNbN0ȏ̊ԁA6502̃vOG~[VɂĎs܂B
	 */
	private final void ExecuteEmulation()
	{
		int clk = CLK;
		int t = T;
		int a = A;
		int[] wram = this.wram;
		int pc = PC;
		final int i0xFF = 0xFF;
		int c = C;
		int x = X;
		int m;
		int y = Y;
		

		while (clk > 0) {
			emulation_loop:
			for (;;) {
				int adr;
				switch (LdMemB_PRG(pc++)) {
					
					// Immediate ADC
					case 0x69:
						m = LdMemB_PRG(pc++);
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 2;
						continue;

					// Zero Page ADC
					case 0x65:
						m = wram[LdMemB_PRG(pc++)];
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 3;
						continue;

					// Zero Page,X ADC
					case 0x75:
						m = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute ADC
					case 0x6D:
						m = LdMemB(LdMemW(pc));
						pc += 2;
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,X ADC
					case 0x7D:
						m = LdMemB(LdMemW(pc) + x);
						pc += 2;
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,Y ADC
					case 0x79:
						m = LdMemB(LdMemW(pc) + y);
						pc += 2;
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Indirect,Y ADC
					case 0x61:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						m = LdMemB(adr);
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 6;
						continue;

					// Indirect,Y ADC ($44),Y
					case 0x71:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						m = LdMemB(adr);
						t = a + m + c;
						c = t >> 8;
						V = (t ^ a) & (t ^ m);
						a = t &= i0xFF;
						clk -= 5;
						continue;

					// Immediate SBC
					case 0xE9:
						m = LdMemB_PRG(pc++);
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 2;
						continue;

					// Zero Page SBC
					case 0xE5:
						m = wram[LdMemB_PRG(pc++)];
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 3;
						continue;

					// Zero Page,X SBC
					case 0xF5:
						m = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute SBC
					case 0xED:
						m = LdMemB(LdMemW(pc));
						pc += 2;
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,X SBC
					case 0xFD:
						m = LdMemB(LdMemW(pc) + x);
						pc += 2;
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,Y SBC
					case 0xF9:
						m = LdMemB(LdMemW(pc) + y);
						pc += 2;
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 4;
						continue;

					// Indirect,X SBC
					case 0xE1:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						m = LdMemB(adr);
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 6;
						continue;

					// Indirect,Y SBC
					case 0xF1:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						m = LdMemB(adr);
						t = a - m - (c ^ 1);
						c = (~t >> 8) & SFLG.C;
						V = (a ^ t) & (a ^ m);
						a = t &= i0xFF;
						clk -= 5;
						continue;

					// Immediate AND
					case 0x29:
						m = LdMemB_PRG(pc++);
						t = a &= m;
						clk -= 2;
						continue;

					// Zero Page AND
					case 0x25:
						m = wram[LdMemB_PRG(pc++)];
						t = a &= m;
						clk -= 3;
						continue;

					// Zero Page,X AND
					case 0x35:
						m = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						t = a &= m;
						clk -= 4;
						continue;

					// Absolute AND
					case 0x2D:
						m = LdMemB(LdMemW(pc));
						pc += 2;
						t = a &= m;
						clk -= 4;
						continue;

					// Absolute,X AND
					case 0x3D:
						m = LdMemB(LdMemW(pc) + x);
						pc += 2;
						t = a &= m;
						clk -= 4;
						continue;

					// Absolute,Y AND
					case 0x39:
						m = LdMemB(LdMemW(pc) + y);
						pc += 2;
						t = a &= m;
						clk -= 4;
						continue;

					// Indirect,X AND
					case 0x21:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						m = LdMemB(adr);
						t = a &= m;
						clk -= 6;
						continue;

					// Indirect,Y AND
					case 0x31:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						m = LdMemB(adr);
						t = a &= m;
						clk -= 5;
						continue;

					// Immediate ORA
					case 0x09:
						m = LdMemB_PRG(pc++);
						t = a |= m;
						clk -= 2;
						continue;

					// Zero Page ORA
					case 0x05:
						m = wram[LdMemB_PRG(pc++)];
						t = a |= m;
						clk -= 3;
						continue;

					// Zero Page,X ORA
					case 0x15:
						m = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						t = a |= m;
						clk -= 4;
						continue;

					// Absolute ORA
					case 0x0D:
						m = LdMemB(LdMemW(pc));
						pc += 2;
						t = a |= m;
						clk -= 4;
						continue;

					// Absolute,X ORA
					case 0x1D:
						m = LdMemB(LdMemW(pc) + x);
						pc += 2;
						t = a |= m;
						clk -= 4;
						continue;

					// Absolute,Y ORA
					case 0x19:
						m = LdMemB(LdMemW(pc) + y);
						pc += 2;
						t = a |= m;
						clk -= 4;
						continue;

					// Indirect,X ORA
					case 0x01:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						m = LdMemB(adr);
						t = a |= m;
						clk -= 6;
						continue;

					// Indirect,Y ORA
					case 0x11:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						m = LdMemB(adr);
						t = a |= m;
						clk -= 5;
						continue;

					// Immediate EOR #$44
					case 0x49:
						m = LdMemB_PRG(pc++);
						t = a ^= m;
						clk -= 2;
						continue;

					// Zero Page EOR
					case 0x45:
						m = wram[LdMemB_PRG(pc++)];
						t = a ^= m;
						clk -= 3;
						continue;

					// Zero Page,X EOR
					case 0x55:
						m = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						t = a ^= m;
						clk -= 4;
						continue;

					// Absolute EOR
					case 0x4D:
						m = LdMemB(LdMemW(pc));
						pc += 2;
						t = a ^= m;
						clk -= 4;
						continue;

					// Absolute,X EOR
					case 0x5D:
						m = LdMemB(LdMemW(pc) + x);
						pc += 2;
						t = a ^= m;
						clk -= 4;
						continue;

					// Absolute,Y EOR
					case 0x59:
						m = LdMemB(LdMemW(pc) + y);
						pc += 2;
						t = a ^= m;
						clk -= 4;
						continue;

					// Indirect,X EOR
					case 0x41:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						m = LdMemB(adr);
						t = a ^= m;
						clk -= 6;
						continue;

					// Indirect,Y EOR
					case 0x51:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						m = LdMemB(adr);
						t = a ^= m;
						clk -= 5;
						continue;

					// Accumulator ASL
					case 0x0A:
						c = a >> 7;
						t = a = (a << 1) & i0xFF;
						clk -= 2;
						continue;

					// Zero Page ASL
					case 0x06:
						adr = LdMemB_PRG(pc++);
						m = LdMemB(adr);
						c = m >> 7;
						t = (m << 1) & i0xFF;
						StMem(adr, t);
						clk -= 5;
						continue;

					// Zero Page,X ASL
					case 0x16:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						m = LdMemB(adr);
						c = m >> 7;
						t = (m << 1) & i0xFF;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute ASL
					case 0x0E:
						adr = LdMemW(pc);
						pc += 2;
						m = LdMemB(adr);
						c = m >> 7;
						t = (m << 1) & i0xFF;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute,X ASL
					case 0x1E:
						adr = LdMemW(pc) + x;
						pc += 2;
						m = LdMemB(adr);
						c = m >> 7;
						t = (m << 1) & i0xFF;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Accumulator LSR
					case 0x4A:
						c = (a & 1);
						t = a >>= 1;
						clk -= 2;
						continue;

					// Zero Page LSR
					case 0x46:
						adr = LdMemB_PRG(pc++);
						m = LdMemB(adr);
						c = (m & 1);
						t = m >> 1;
						StMem(adr, t);
						clk -= 5;
						continue;

					// Zero Page,X LSR
					case 0x56:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						m = LdMemB(adr);
						c = (m & 1);
						t = m >> 1;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute LSR $4400
					case 0x4E:
						adr = LdMemW(pc);
						pc += 2;
						m = LdMemB(adr);
						c = (m & 1);
						t = m >> 1;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute,X LSR
					case 0x5E:
						adr = LdMemW(pc) + x;
						pc += 2;
						m = LdMemB(adr);
						c = (m & 1);
						t = m >> 1;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Accumulator ROL
					case 0x2A:
						m = a;
						t = a = ((a << 1) | c) & i0xFF;
						c = m >> 7;
						clk -= 2;
						continue;

					// Zero Page ROL
					case 0x26:
						adr = LdMemB_PRG(pc++);
						m = LdMemB(adr);
						t = ((m << 1) | c) & i0xFF;
						c = m >> 7;
						StMem(adr, t);
						clk -= 5;
						continue;

					// Zero Page,X ROL
					case 0x36:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						m = LdMemB(adr);
						t = ((m << 1) | c) & i0xFF;
						c = m >> 7;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute ROL
					case 0x2E:
						adr = LdMemW(pc);
						pc += 2;
						m = LdMemB(adr);
						t = ((m << 1) | c) & i0xFF;
						c = m >> 7;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute,X ROL
					case 0x3E:
						adr = LdMemW(pc) + x;
						pc += 2;
						m = LdMemB(adr);
						t = ((m << 1) | c) & i0xFF;
						c = m >> 7;
						StMem(adr, t);
						clk -= 6;
						continue;

					// Accumulator ROR
					case 0x6A:
						m = a;
						t = a = (a >> 1) | (c << 7);
						c = (m & 1);
						clk -= 2;
						continue;

					// Zero Page ROR
					case 0x66:
						adr = LdMemB_PRG(pc++);
						m = LdMemB(adr);
						t = (m >> 1) | (c << 7);
						c = (m & 1);
						StMem(adr, t);
						clk -= 5;
						continue;

					// Zero Page,X ROR
					case 0x76:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						m = LdMemB(adr);
						t = (m >> 1) | (c << 7);
						c = (m & 1);
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute ROR
					case 0x6E:
						adr = LdMemW(pc);
						pc += 2;
						m = LdMemB(adr);
						t = (m >> 1) | (c << 7);
						c = (m & 1);
						StMem(adr, t);
						clk -= 6;
						continue;

					// Absolute,X ROR
					case 0x7E:
						adr = LdMemW(pc) + x;
						pc += 2;
						m = LdMemB(adr);
						t = (m >> 1) | (c << 7);
						c = (m & 1);
						StMem(adr, t);
						clk -= 6;
						continue;

					// Relative BCC
					case 0x90:
						if (c == 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BCS
					case 0xB0:
						if (c != 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BEQ
					case 0xF0:
						if (t <= 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BNE
					case 0xD0:
						if (t > 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BVC
					case 0x50:
						if ((V & 0x80) == 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BVS
					case 0x70:
						if ((V & 0x80) != 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BPL
					case 0x10:
						if ((t & 0x80) == 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Relative BMI
					case 0x30:
						if ((t & 0x80) != 0) {
							clk -= 3;
							pc += (byte) LdMemB_PRG(pc) + 1;
							break emulation_loop;
						}
						pc++;
						clk -= 2;
						continue;

					// Zero Page BIT
					case 0x24:
						t = wram[LdMemB_PRG(pc++)];
						V = t << 1;
						if ((a & t) == 0)
							t |= ~0x1FF;
						clk -= 3;
						continue;

					// Absolute BIT
					case 0x2C:
						t = LdMemB(LdMemW(pc));
						pc += 2;
						V = t << 1;
						if ((a & t) == 0)
							t |= ~0x1FF;
						clk -= 4;
						continue;

					// Absolute JMP
					case 0x4C:
						pc = LdMemW(pc);
						clk -= 3;
						break emulation_loop;

					// Indirect JMP
					case 0x6C:
						adr = LdMemW(pc);
						pc = LdMemW(adr);
						clk -= 5;
						break emulation_loop;

					// Absolute JSR
					case 0x20:
						adr = LdMemW(pc);
						pc += 2;
						PushStackW(pc - 1);
						pc = adr;
						clk -= 6;
						break emulation_loop;

					// Implied RTS
					case 0x60:
						pc = PullStackW() + 1;
						clk -= 6;
						break emulation_loop;

					// Implied BRK
					case 0x00:
						clk -= 7;
						if ((I & SFLG.I) != 0) {
							continue;
						}

						B = SFLG.B;
						PushStackW(pc + 1);
						PushStackB(c | ((t <= 0) ? SFLG.Z : 0) | I | D | B | R
								| (((V & 0x80) != 0) ? SFLG.V : 0)
								| (((t & 0x80) != 0) ? SFLG.N : 0));
						I = SFLG.I;
						pc = LdMemW(0xfffe);
						break emulation_loop;

					// Implied RTI
					case 0x40:
						m = PullStackB();
						c = m & SFLG.C;
						I = m & SFLG.I;
						D = m & SFLG.D;
						B = m & SFLG.B;
						R = m & SFLG.R;
						V = m << 1;
						t = ((m & 0x82) ^ 2) | ((m & 2) << 30);
						pc = PullStackW();
						clk -= 6;
						break emulation_loop;

					// Immediate CMP
					case 0xC9:
						t = a - LdMemB_PRG(pc++);
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 2;
						continue;

					// Zero Page CMP
					case 0xC5:
						t = a - wram[LdMemB_PRG(pc++)];
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 3;
						continue;

					// Zero Page,X CMP
					case 0xD5:
						t = a - wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute CMP
					case 0xCD:
						t = a - LdMemB(LdMemW(pc));
						pc += 2;
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,X CMP
					case 0xDD:
						t = a - LdMemB(LdMemW(pc) + x);
						pc += 2;
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Absolute,Y CMP
					case 0xD9:
						t = a - LdMemB(LdMemW(pc) + y);
						pc += 2;
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Indirect,X CMP
					case 0xC1:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						t = a - LdMemB(adr);
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 6;
						continue;

					// Indirect,Y CMP
					case 0xD1:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						t = a - LdMemB(adr);
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 5;
						continue;

					// Immediate CPX
					case 0xE0:
						t = x - LdMemB_PRG(pc++);
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 2;
						continue;

					// Zero Page CPX
					case 0xE4:
						t = x - wram[LdMemB_PRG(pc++)];
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 3;
						continue;

					// Absolute CPX
					case 0xEC:
						t = x - LdMemB(LdMemW(pc));
						pc += 2;
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Immediate CPY
					case 0xC0:
						t = y - LdMemB_PRG(pc++);
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 2;
						continue;

					// Zero Page CPY	
					case 0xC4:
						t = y - wram[LdMemB_PRG(pc++)];
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 3;
						continue;

					// Absolute CPY
					case 0xCC:
						t = y - LdMemB(LdMemW(pc));
						pc += 2;
						c = (~t >> 8) & 1;
						t &= i0xFF;
						clk -= 4;
						continue;

					// Zero Page INC
					case 0xE6:
						adr = LdMemB_PRG(pc++);
						StMem(adr, t = (LdMemB(adr) + 1) & i0xFF);
						clk -= 5;
						continue;

					// Zero Page,X INC
					case 0xF6:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						StMem(adr, t = (LdMemB(adr) + 1) & i0xFF);
						clk -= 6;
						continue;

					// Absolute INC
					case 0xEE:
						adr = LdMemW(pc);
						pc += 2;
						StMem(adr, t = (LdMemB(adr) + 1) & i0xFF);
						clk -= 6;
						continue;

					// Absolute,X INC
					case 0xFE:
						adr = LdMemW(pc) + x;
						pc += 2;
						StMem(adr, t = (LdMemB(adr) + 1) & i0xFF);
						clk -= 6;
						continue;

					// Zero Page DEC
					case 0xC6:
						adr = LdMemB_PRG(pc++);
						StMem(adr, t = (LdMemB(adr) - 1) & i0xFF);
						clk -= 5;
						continue;

					// Zero Page,X DEC
					case 0xD6:
						adr = (LdMemB_PRG(pc++) + x) & i0xFF;
						StMem(adr, t = (LdMemB(adr) - 1) & i0xFF);
						clk -= 6;
						continue;

					// Absolute DEC $4400
					case 0xCE:
						adr = LdMemW(pc);
						pc += 2;
						StMem(adr, t = (LdMemB(adr) - 1) & i0xFF);
						clk -= 6;
						continue;

					// Absolute,X DEC
					case 0xDE:
						adr = LdMemW(pc) + x;
						pc += 2;
						StMem(adr, t = (LdMemB(adr) - 1) & i0xFF);
						clk -= 6;
						continue;

					// Implied INX
					case 0xE8:
						t = x = (x + 1) & i0xFF;
						clk -= 2;
						continue;

					// Implied DEX
					case 0xCA:
						t = x = (x - 1) & i0xFF;
						clk -= 2;
						continue;

					// Implied INY
					case 0xC8:
						t = y = (y + 1) & i0xFF;
						clk -= 2;
						continue;

					// Implied DEY
					case 0x88:
						t = y = (y - 1) & i0xFF;
						clk -= 2;
						continue;

					// Implied CLC
					case 0x18:
						c = 0;
						clk -= 2;
						continue;

					// Implied SEC
					case 0x38:
						c = 1;
						clk -= 2;
						continue;

					// Implied CLI
					case 0x58:
						I = 0;
						clk -= 2;
						continue;

					// Implied SEI
					case 0x78:
						I = SFLG.I;
						clk -= 2;
						continue;

					// Implied CLD $D8
					case 0xD8:
						D = 0;
						clk -= 2;
						continue;

					// Implied SED $F8
					case 0xF8:
						D = SFLG.D;
						clk -= 2;
						continue;

					// Implied CLV $B8
					case 0xB8:
						V = 0;
						clk -= 2;
						continue;

					// Immediate LDA
					case 0xA9:
						t = a = LdMemB_PRG(pc++);
						clk -= 2;
						continue;

					// Zero Page LDA
					case 0xA5:
						t = a = wram[LdMemB_PRG(pc++)];
						clk -= 3;
						continue;

					// Zero Page,X LDA
					case 0xB5:
						t = a = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						clk -= 4;
						continue;

					// Absolute LDA
					case 0xAD:
						t = a = LdMemB(LdMemW(pc));
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,X LDA
					case 0xBD:
						t = a = LdMemB(LdMemW(pc) + x);
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,Y LDA
					case 0xB9:
						t = a = LdMemB(LdMemW(pc) + y);
						pc += 2;
						clk -= 4;
						continue;

					// Indirect,X LDA
					case 0xA1:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						t = a = LdMemB(adr);
						clk -= 6;
						continue;

					// Indirect,Y LDA
					case 0xB1:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						t = a = LdMemB(adr);
						clk -= 5;
						continue;

					// Immediate LDX
					case 0xA2:
						t = x = LdMemB_PRG(pc++);
						clk -= 2;
						continue;

					// Zero Page LDX
					case 0xA6:
						t = x = wram[LdMemB_PRG(pc++)];
						clk -= 3;
						continue;

					// Zero Page,Y LDX
					case 0xB6:
						t = x = LdMemB((LdMemB_PRG(pc++) + y) & i0xFF);
						clk -= 4;
						continue;

					// Absolute LDX
					case 0xAE:
						t = x = LdMemB(LdMemW(pc));
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,Y LDX
					case 0xBE:
						t = x = LdMemB(LdMemW(pc) + y);
						pc += 2;
						clk -= 4;
						continue;

					// Immediate LDY
					case 0xA0:
						t = y = LdMemB_PRG(pc++);
						clk -= 2;
						continue;

					// Zero Page LDY
					case 0xA4:
						t = y = wram[LdMemB_PRG(pc++)];
						clk -= 3;
						continue;

					// Zero Page,X LDY
					case 0xB4:
						t = y = wram[(LdMemB_PRG(pc++) + x) & i0xFF];
						clk -= 4;
						continue;

					// Absolute LDY
					case 0xAC:
						t = y = LdMemB(LdMemW(pc));
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,X LDY
					case 0xBC:
						t = y = LdMemB(LdMemW(pc) + x);
						pc += 2;
						clk -= 4;
						continue;

					// Zero Page STA
					case 0x85:
						StMem(LdMemB_PRG(pc++), a);
						clk -= 3;
						continue;

					// Zero Page,X
					case 0x95:
						StMem((LdMemB_PRG(pc++) + x) & i0xFF, a);
						clk -= 4;
						continue;

					// Absolute STA
					case 0x8D:
						StMem(LdMemW(pc), a);
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,X STA
					case 0x9D:
						StMem(LdMemW(pc) + x, a);
						pc += 2;
						clk -= 4;
						continue;

					// Absolute,Y STA $4400,Y
					case 0x99:
						StMem(LdMemW(pc) + y, a);
						pc += 2;
						clk -= 4;
						continue;

					// Indirect,X STA ($44,X)
					case 0x81:
						adr = LdMemB_PRG(pc++) + x;
						adr = wram[adr & i0xFF] | (wram[(adr + 1) & i0xFF] << 8);
						StMem(adr, a);
						clk -= 6;
						continue;

					// Indirect,Y STA ($44),Y
					case 0x91:
						adr = LdMemB_PRG(pc++);
						adr = (wram[adr] | (wram[(adr + 1) & i0xFF] << 8)) + y;
						StMem(adr, a);
						clk -= 5;
						continue;

					// Zero Page STX
					case 0x86:
						StMem(LdMemB_PRG(pc++), x);
						clk -= 3;
						continue;

					// Zero Page,Y STX
					case 0x96:
						StMem((LdMemB_PRG(pc++) + y) & i0xFF, x);
						clk -= 4;
						continue;

					// Absolute STX
					case 0x8E:
						StMem(LdMemW(pc), x);
						pc += 2;
						clk -= 4;
						continue;

					// Zero Page STY
					case 0x84:
						StMem(LdMemB_PRG(pc++), y);
						clk -= 3;
						continue;

					// Zero Page,X STY
					case 0x94:
						StMem((LdMemB_PRG(pc++) + x) & i0xFF, y);
						clk -= 4;
						continue;

					// Absolute STY
					case 0x8C:
						StMem(LdMemW(pc), y);
						pc += 2;
						clk -= 4;
						continue;

					// Implied TAX
					case 0xAA:
						t = x = a;
						clk -= 2;
						continue;

					// Implied TXA
					case 0x8A:
						t = a = x;
						clk -= 2;
						continue;

					// Implied TAY
					case 0xA8:
						t = y = a;
						clk -= 2;
						continue;

					// Implied TYA
					case 0x98:
						t = a = y;
						clk -= 2;
						continue;

					// Implied TXS
					case 0x9A:
						SP = x;
						clk -= 2;
						continue;

					// Implied TSX
					case 0xBA:
						t = x = SP;
						clk -= 2;
						continue;

					// Implied PHA $48
					case 0x48:
						PushStackB(a);
						clk -= 3;
						continue;

					// Implied PLA $68
					case 0x68:
						t = a = PullStackB();
						clk -= 4;
						continue;

					// Implied PHP $08
					case 0x08:
						PushStackB(c | ((t <= 0) ? SFLG.Z : 0) | I | D | B | R
								| (((V & 0x80) != 0) ? SFLG.V : 0)
								| (((t & 0x80) != 0) ? SFLG.N : 0));
						clk -= 3;
						continue;

					// Implied PLP
					case 0x28:
						m = PullStackB();
						c = m & SFLG.C;
						I = m & SFLG.I;
						D = m & SFLG.D;
						B = m & SFLG.B;
						R = m & SFLG.R;
						V = m << 1;
						t = ((m & 0x82) ^ 2) | ((m & 2) << 30);
						clk -= 4;
						continue;

					// Implied NOP
					case 0xEA:
						clk -= 2;
						continue;
				}
			}

			clk -= DMAclk;
			DMAclk = 0;
		}
		
		CLK = clk - DMAclk;
		DMAclk = 0;
		T = t;
		A = a;
		PC = pc;
		C = c;
		X = x;
		Y = y;
	}
	
	/*REMOVE AND INSERT TRANSLATED CODE*/
	
	public static void LoadRom(NESFile nesFile)
	{
		nesFile.loadRom("SuperMarioBros.nes");
	}
	
	/**
	 * FOMAJ[hԍ擾܂B
	 * @return
	 */
	protected final String getUSER_ID()
	{
		return "123456789ABCDEF1231F";
	}

	/**
	 * csNbN0傫ԁA6502RA쓮āAvOs܂B
	 */
	public final void ExecuteFast()
	{		
		ExecuteEmulation();
	}

}
