import java.io.IOException;
import java.io.InputStream;

import javax.microedition.io.Connector;

public class NESFile
{
	private int[] prgRom;
	private int[] chrRom;
	private byte prgPageSize;
	private byte chrPageSize;
	private byte[] settings;
	
	public NESFile()
	{
	}
	
	public int getPrgSize()
	{
		return prgPageSize * 0x4000;
	}

	public int getChrSize()
	{
		return chrPageSize * 0x2000;
	}

	public boolean isHorizontalMirror()
	{
		return (settings[0] & 1) == 0;
	}

	public boolean isBatteryBackup()
	{
		return (settings[0] & 2) == 1;
	}

	public boolean isTrainer()
	{
		return (settings[0] & 4) == 1;
	}

	public boolean isOneScreen()
	{
		return (settings[0] & 8) == 1;
	}

	public int getMapper()
	{
		return ((settings[1] & 0xF0) << 4) | (settings[0] & 0xF0);
	}

	public int[] getPrgRom()
	{
		return prgRom;
	}

	public int[] getChrRom()
	{
		return chrRom;
	}
	
	public void loadRom(String fileName)
	{
		InputStream is = null;
		
		try	{
			is = Connector.openInputStream("resource:///" + fileName);
			
			byte[] id = new byte[4];
			is.read(id);
			
			if (id[0] != 0x4E
					|| id[1] != 0x45
					|| id[2] != 0x53
					|| id[3] != 0x1A) {
					throw new IOException();
			}
			
			prgPageSize = (byte)is.read();
			chrPageSize = (byte)is.read();
			
			settings = new byte[2];
			is.read(settings);
			
			is.skip(8);
			
			prgRom = new int[getPrgSize()];
			byte[] tmp = new byte[getPrgSize()];			
			is.read(tmp);
			for (int i = 0; i < tmp.length; i++) {
				prgRom[i] = tmp[i] & 0xff;
			}
			
			chrRom = new int[getChrSize()];
			tmp = new byte[getChrSize()];
			is.read(tmp);
			for (int i = 0; i < tmp.length; i++) {
				chrRom[i] = tmp[i] & 0xff;
			}
		}
		catch (Exception e) {
		}
		finally {
			if (is != null) {
				try	{
					is.close();
				}
				catch (IOException e) {
				}
			}
		}
	}
}
