import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import com.nttdocomo.ui.Graphics;
import com.nttdocomo.ui.Image;

public class PPU {
	
	private final int[] chrRom;

	private final int[] chrRom32;

	private final int[] vram;

	private final int[] palette;
	
	private final int[] paletteBack;

	private final int[] spriteRam;

	private final int[] spriteRamBack;

	private int controlRegister1;

	private int controlRegister2;

	private int statusRegister;

	private int spriteMemoryAddress;

	private int bgScrollOffsetH;

	private int bgScrollOffsetV;

	private boolean bgScrollRatch;
	
	private boolean ppuMemoryAddressRatch;

	private int ppuMemoryAddress;

	private boolean isPpuReadFirst = true;
	
	private final int nameTblMask;

	private final int[] buffer;
	
	private final boolean[] isDirty;
	
	public boolean isDirtyAll = false;

	private int scanLine;
	
	private final Image[] spriteImg;
	
	private final Graphics[] spriteGraphics;
	
	private final Image nameTblImage;
	
	private final Graphics nameTblGraphics;

	private final boolean[] isTransparentOnly;

	private final int[] paletteData = new int[] { 
			0x757575, 0x271B8F, 0x0000AB, 0x47009F, 0x8F0077, 0xAB0013, 0xA70000, 0x7F0B00, 0x432F00,
			0x004700, 0x005100, 0x003F17, 0x1B3F5F, 0x000000, 0x050505, 0x050505, 0xBCBCBC, 0x0073EF,
			0x233BEF, 0x8300F3, 0xBF00BF, 0xE7005B, 0xDB2B00, 0xCB4F0F, 0x8B7300, 0x009700, 0x00AB00,
			0x00933B, 0x00838B, 0x111111, 0x090909, 0x090909, 0xFFFFFF, 0x3FBFFF, 0x5F97FF, 0xA78BFD,
			0xF77BFF, 0xFF77B7, 0xFF7763, 0xFF9B3B, 0xF3BF3F, 0x83D313, 0x4FDF4B, 0x58F898, 0x00EBDB,
			0x666666, 0x0D0D0D, 0x0D0D0D, 0xFFFFFF, 0xABE7FF, 0xC7D7FF, 0xD7CBFF, 0xFFC7FF, 0xFFC7DB,
			0xFFBFB3, 0xFFDBAB, 0xFFE7A3,0xE3FFA3, 0xABF3BF, 0xB3FFCF, 0x9FFFF3, 0xDDDDDD, 0x111111,
			0x111111 };

	private final int[][] pal2rgbTbl;
		
	private int[] pal2rgb;
	
	private int bgChrRom32Index;

	private final int[] scrollHBuffer;
	
	private final int[] scrollVBuffer;

	private final int[] scrollHBufferBack;
	
	private final int[] scrollVBufferBack;

	private final int[] ppuControlReg1Buffer;
	
	private final int[] ppuControlReg1BufferBack;

	protected final int SCREEN_OFFSET_X;

	protected final int SCREEN_OFFSET_Y;
	
	private final NESFile nesFile;
	
	private int vramMirrorMask;
	
	boolean[] isDirtyPaletteSet;
	
	public final int SCAN_LINE_COUNT = 262;
	
	PPU(NESFile nesFile, int screenWidth, int screenHeight) {
		
		// NESFile
		this.nesFile = nesFile;
		
		// PPU
		this.chrRom = nesFile.getChrRom();
		this.chrRom32 = new int[chrRom.length >> 2];
		this.vram = new int[0x1000];
		this.palette = new int[0x0020];
		this.paletteBack = new int[0x0020];
		this.pal2rgbTbl = new int[8][4];
		this.spriteRam = new int[0x100];
		this.spriteRamBack = new int[0x100];
		this.statusRegister = 0;
		this.bgScrollRatch = true;
		this.ppuMemoryAddressRatch = true;
		this.buffer = new int[64];
		this.spriteImg = new Image[64];
		this.spriteGraphics = new Graphics[64];
		this.isTransparentOnly = new boolean[chrRom.length >> 4];
		this.nameTblImage = Image.createImage(512, 480);
		this.nameTblGraphics = nameTblImage.getGraphics();	
		this.isDirty = new boolean[32 * 32 * 4];
		for (int i = 0; i < isDirty.length; i++) {
			isDirty[i] = true;
		}
		
		for (int i = 0; i < 64; i++) {
			spriteImg[i] = Image.createImage(8, 16);
			spriteGraphics[i] = spriteImg[i].getGraphics();
		}

		makeChrRom32();
		setPaletteData();
		makeTransparentOnlyChrTbl();

		this.isDirtyPaletteSet = new boolean[8];
		
		this.scrollHBuffer = new int[SCAN_LINE_COUNT];
		this.scrollVBuffer = new int[SCAN_LINE_COUNT];
		this.ppuControlReg1Buffer = new int[SCAN_LINE_COUNT];

		this.scrollHBufferBack = new int[SCAN_LINE_COUNT];
		this.scrollVBufferBack = new int[SCAN_LINE_COUNT];
		this.ppuControlReg1BufferBack = new int[SCAN_LINE_COUNT];
				
		this.nameTblMask = nesFile.isHorizontalMirror() ? 0xBFF : 0x7FF; // 4-screen0xFFFŃ}XN		
					
		this.SCREEN_OFFSET_X = (screenWidth - 256) / 2;
		this.SCREEN_OFFSET_Y = (screenHeight - 240) / 2;
			
		// Mirrormask
		if (nesFile.isHorizontalMirror()) {
			vramMirrorMask = 2;
		} else {
			vramMirrorMask = 1;
		}
	}
	
	public final int getScanLine() {
		return this.scanLine;
	}
	
	public final void setScanLine(int val) {
		this.scanLine = val;
	}

	/**
	 * NES̃pbgf[^zRGBlAgщʂ̐F[x̐[ɂ킹ăXP[
	 * lōĐݒ肵܂B
	 */
	private final void setPaletteData() {
		for (int i = 0; i < paletteData.length; i++) {
			int rgb = paletteData[i];
			paletteData[i] = Graphics.getColorOfRGB((rgb >> 16) & 0xff,
					(rgb >> 8) & 0xff, rgb & 0xff);
		}
	}

	protected final void ClearPPUWriteBuffer()
	{
		for (int i = 0; i < ppuControlReg1Buffer.length; i++) {
			ppuControlReg1Buffer[i] = -1;
			scrollHBuffer[i] = -1;
			scrollVBuffer[i] = -1;
		}
	}
	
	protected final void StorePaletteBack()
	{
		for (int i = 0; i < palette.length; i++) {
			paletteBack[i] = palette[i];
		}
	}
		
	/**
	 * CHR-ROMC[W̓e`揈s₷悤ɕϊAchrRom32z܂B
	 */
	private final void makeChrRom32() {
		int index = 0;
		int index32 = 0;
		for (int chrNo = 0; chrNo < (chrRom.length / 16); chrNo++) { // 161chrbyte
			for (int i = 0; i < 4; i++) {
				int pat = 0;
				for (int j = 0; j < 2; j++) {
					int patLow = chrRom[index];
					int patHigh = chrRom[index + 8] << 1;
					index++;
					for (int k = 0; k < 8; k++) {
						pat <<= 2;
						pat |= (patHigh & 2) | (patLow & 1);
						patHigh >>= 1;
						patLow >>= 1;
					}
				}
				pat = ((pat >> 16) & 0xFFFF) | (pat << 16);
				chrRom32[index32] = pat;
				index32++;
			}
			index += 8;
		}
	}
	
	/**
	 * CHR-ROMC[ŴAʉߐFx^̃LN^𔻕ʂ邽߂̃tOr܂B
	 */
	private final void makeTransparentOnlyChrTbl() {
		for (int i = 0; i < chrRom.length; i += 16) {
			int sum = 0;
			for (int j = i; j < i + 16; j++) {
				sum += chrRom[j];
			}
			isTransparentOnly[i >> 4] = (sum == 0);
		}
	}

	/**
	 * w肵XLCŁAXvCg[qbgĂǂ𒲂ׂ܂B
	 * @param scanline
	 */
	public final void CheckSpriteZeroHit()
	{
		int sprY = spriteRam[0];
//		if (scanLine >= sprY && scanLine < sprY + 8) {
		if (scanLine == sprY + 8) {
				SetSpriteZeroHit();
		}
	}
	
	/**
	 * XvCg[qbg̏Ԃɂ܂B
	 */
	public final void SetSpriteZeroHit() {
		statusRegister |= (1 << 6);
	}

	/**
	 * XvCg[qbgł͂ȂԂɂ܂B
	 */
	public final void ClearSpriteZeroHit() {
		statusRegister &= ~(1 << 6);
	}

	/**
	 * uLOԂ̃tOZbg܂B
	 */
	public final void SetVBlank() {
		statusRegister |= 1 << 7;
		bgScrollRatch = true;
		// if ((controlRegister1 & (1 << 7)) != 0)
		// {
		// NMI();
		// }
	}

	/**
	 * uLOԂ̃tONA܂B
	 */
	public final void ClearVBlank() {
		statusRegister &= 0x7F;
	}
	
	/**
	 * NMI𔭐܂B
	 */
	public final void TrigNMI(NES nes) {
		if ((controlRegister1 & (1 << 7)) != 0) {
			nes.NMI();
		}
	}

	/**
	 * w肵AhXPPUԂA1oCgǂݍ݂܂B
	 * @param address
	 * @return
	 */
	private final int LdPPU(int address) {
		// if (address <= 0x0FFF)
		// {
		// // p^[e[u#0
		// return chrRom[address];
		// } else
		if (address <= 0x1FFF) {
			// p^[e[u#0,#1
			return chrRom[address];
		} else if (address <= 0x3EFF) {
			// l[e[u
			// e[u
			// yу~[
			return vram[address & nameTblMask];
		} else if (address <= 0x3FFF) {
			// obNOEhpbg
			// XvCgpbg
			// yу~[
			return palette[address & 0x1F];
		}

		return 0;
	}

	/**
	 * w肵AhXbPPUԂɁA1oCg݂܂B
	 * @param address
	 * @param val
	 */
	private final void StPPU(int address, int val) {
		// if (address <= 0x0FFF)
		// {
		// // p^[e[u#0
		// }
		// else if (address <= 0x1FFF)
		// {
		// // p^[e[u#1
		// }
		if (address <= 0x1FFF) {
		} else 	if (address <= 0x3EFF) {
			// l[e[u
			// e[u
			// yу~[
			vram[address & nameTblMask] = val;

			// $2000` $03C0 l[e[u
			// $23C0` $0040 e[u
			// $2400` $03C0 l[e[u
			// $27C0` $0040 e[u
			// $2800` $03C0 l[e[u
			// $2BC0` $0040 e[u
			// $2C00` $03C0 l[e[u
			// $2FC0` $0040 e[u
			if ((address & 0x3FF) < 0x3C0) {
				// l[e[u
				isDirty[address & nameTblMask] = true;
			} else { //if ((address & 0x3FF) < 0x4B0) { // 0x4B0(32*30)+(16*15)
				// e[u
		//		int tblNo = ((address & nameTblMask) >> 10) & 2;
		//		int chrNo = ((address & 0x3FF) - 0x3C0) << 2;
		//		int offset = tblNo << 10; // 1024{
				int offset = (address & nameTblMask);
				int attX = offset & 7;
				int attY = (offset >> 3) & 7;
				offset = (offset & (3 << 10)) + (attY << 7) + (attX << 2);
						
				isDirty[offset    ] = true;
				isDirty[offset + 1] = true;
				isDirty[offset + 2] = true;
				isDirty[offset + 3] = true;
				isDirty[offset + 32] = true;
				isDirty[offset + 33] = true;
				isDirty[offset + 34] = true;
				isDirty[offset + 35] = true;
				isDirty[offset + 64] = true;
				isDirty[offset + 65] = true;
				isDirty[offset + 66] = true;
				isDirty[offset + 67] = true;
				isDirty[offset + 96] = true;
				isDirty[offset + 97] = true;
				isDirty[offset + 98] = true;
				isDirty[offset + 99] = true;
			}
		} else if (address <= 0x3FFF) {
			// obNOEhpbg
			// XvCgpbg
			// yу~[
			if ((address & 0xF) == 0)
			{
				int offset = address & 0x10;
				// $3F00ɏ邷ׂẴf[^́A$3F04A$3F08A$3F0CɃ~[ 
				// $3F10ɏ邷ׂẴf[^́A$3F14A$3F18A$3F1CɃ~[
				palette[offset + 0xC] = palette[offset + 0x8] = palette[offset + 0x4] = palette[offset] = val;
			} else {
				palette[address & 0x1F] = val;
			}
		}
	}

	/**
	 * PPURg[WX^1̒lݒ肵܂B
	 * @param val
	 */
	protected final void SetControlRegister1(int val) {
		controlRegister1 = val;
		ppuControlReg1Buffer[scanLine] = val;
	}

	/**
	 * PPURg[WX^2̒lݒ肵܂B
	 * @param val
	 */
	protected final void SetControlRegister2(int val) {
		controlRegister2 = val;
	}

	/**
	 * PPUXe[^XWX^̒l擾܂B
	 * @return
	 */
	public final int GetStatusRegister() {
		bgScrollRatch = true;
		ppuMemoryAddressRatch = true;
		int ret = statusRegister;
		statusRegister &= 0x3F;
		return ret;
	}

	/**
	 * XvCgAhXݒ肵܂B
	 * @param val
	 */
	protected final void SetSpriteMemoryAddress(int val) {
		spriteMemoryAddress = val;
	}

	/**
	 * XvCgɒl݂܂B
	 * @param val
	 */
	protected final void SetSpriteMemoryData(int val) {
		spriteRam[spriteMemoryAddress] = val;
		spriteMemoryAddress = (spriteMemoryAddress + 1) & 0xff;
	}

	/**
	 * XN[WX^̒lݒ肵܂B
	 * @param val
	 */
	protected final void SetBGScrollOffset(int val) {
		if (bgScrollRatch) {
			bgScrollOffsetH = val;
			scrollHBuffer[scanLine] = val;
		} else {
			bgScrollOffsetV = (val > 239) ? 0 : val;
			scrollVBuffer[scanLine] = bgScrollOffsetV;

		}
		bgScrollRatch = !bgScrollRatch;
	}

	/**
	 * PPUAhXݒ肵܂B
	 * @param val
	 */
	protected final void SetPPUMemoryAddress(int val) {
		if (ppuMemoryAddressRatch)
		{
			ppuMemoryAddress = (ppuMemoryAddress & 0x00FF) | (val << 8);
		}
		else
		{
			ppuMemoryAddress = (ppuMemoryAddress & 0xFF00) | val;
		}
		
		ppuMemoryAddressRatch = !ppuMemoryAddressRatch;
		isPpuReadFirst = true;
	}

	/**
	 * PPU1oCgǂݍ݂܂B
	 * @return
	 */
	protected final int GetPPUMemoryData() {
		if (isPpuReadFirst) {
			isPpuReadFirst = false;
			return 0;
		}

		int ret = LdPPU(ppuMemoryAddress);
		// ppuMemoryAddress = (ppuMemoryAddress + 1 + (31 & ((controlRegister1 &
		// 3) - 1))) & 0xffff;

		if ((controlRegister1 & (1 << 2)) == 0) {
			ppuMemoryAddress = (ppuMemoryAddress + 1) & 0xffff;
		} else {
			ppuMemoryAddress = (ppuMemoryAddress + 32) & 0xffff;
		}
		return ret;
	}

	/**
	 * PPU1oCg݂܂B
	 * @param val
	 */
	protected final void SetPPUMemoryData(int val) {
		StPPU(ppuMemoryAddress, val);
		// ppuMemoryAddress = (ppuMemoryAddress + 1 + (31 & ((controlRegister1 &
		// 3) - 1))) & 0xffff;
		if ((controlRegister1 & (1 << 2)) == 0) {
			ppuMemoryAddress = (ppuMemoryAddress + 1) & 0xffff;
		} else {
			ppuMemoryAddress = (ppuMemoryAddress + 32) & 0xffff;
		}
	}

	/**
	 * DMA𔭐āACPUԂAXvCgRAMփf[^]s܂B
	 * @param address
	 */
	protected final void DMA(NES nes, int address) {
		if (address <= 0x1FFF) {
			// WRAMSpriteRAMɓ]
			address = (address << 8) & 0x7FF;
			System.arraycopy(nes.wram, address, spriteRam, 0, spriteRam.length);
		} else if (address > 0x7FFF) {
			// PRG-ROMSpriteRAMɓ]
			address = (address << 8) & nes.prgMask;
			System.arraycopy(nes.prgRom, address, spriteRam, 0, spriteRam.length);
		} else {
			// ȊO
			address <<= 8;
			for (int i = 0; i < 256;) {
				spriteRam[i] = nes.LdMemB(address + i);
				spriteRam[i + 1] = nes.LdMemB(address + i + 1);
				spriteRam[i + 2] = nes.LdMemB(address + i + 2);
				spriteRam[i + 3] = nes.LdMemB(address + i + 3);
				i += 4;
			}
		}

		nes.DMAclk += 512;
	}
	
//	private Graphics g;
			
	/**
	 * 1t[̉ʃC[W_O܂B
	 * @param g
	 */
	public final void Render(Graphics g) {
		makePal2rgb();
		fillScrollBuffer();
				
		// bit3 wiL 0:A1:L
		if ((controlRegister2 & (1 << 3)) != 0) {
			//this.g = g;
			if (isDirtyAll) {
				DrawBGToNameTblImageAll();				
			} else {
				DrawBGToNameTblImage();
			}
			DrawNameTblImageToScreen(g);
		}
		
		// bit4 XvCgL 0:A1:L
		if ((controlRegister2 & (1 << 4)) != 0) {
			
			// bit5 XvCgTCY@0:8x8A1:8x16
			if ((controlRegister1 & (1 << 5)) == 0) {
				DrawSprite8(g);				
			} else {
				DrawSprite16(g);				
			}
		}

		// XvCgdouble buffering
		System.arraycopy(spriteRam, 0, spriteRamBack, 0, spriteRamBack.length);
		
		isDirtyAll = false;
	}

	/**
	 * NESpbglARGBl֕ϊ邽߂̃e[u܂B
	 */
	private final void makePal2rgb() {
		for (int palSet = 0; palSet < pal2rgbTbl.length; palSet++) {
			for (int palNo = 0; palNo < pal2rgbTbl[palSet].length; palNo++) {
				pal2rgbTbl[palSet][palNo] = paletteData[palette[(palSet << 2) + palNo]];
			}			
		}
		
		for (int palSet = 0; palSet < 4; palSet++) {
			pal2rgbTbl[palSet][0] = pal2rgbTbl[4][0];
		}
		
		for (int i = 0; i < isDirtyPaletteSet.length; i++) {
			isDirtyPaletteSet[i] = false;
		}
		for (int i = 0; i < palette.length; i++) {
			isDirtyPaletteSet[i >> 2] |=  (palette[i] != paletteBack[i]);
		}
	}
	
	/**
	 * XN[obt@̒l܂ĂȂ𖄂߂܂B
	 */
	private final void fillScrollBuffer() {
		int scroll = scrollHBufferBack[scrollHBufferBack.length - 1];
		for (int i = 0; i < scrollHBuffer.length; i++) {
			if (scrollHBuffer[i] >= 0) {
				scroll = scrollHBuffer[i];
			}
			scrollHBuffer[i] = scroll;
		}
		
		scroll = scrollVBufferBack[scrollVBufferBack.length - 1];
		for (int i = 0; i < scrollVBuffer.length; i++) {
			if (scrollVBuffer[i] >= 0) {
				scroll = scrollVBuffer[i];
			}
			scrollVBuffer[i] = scroll;
		}
		
		int ctrl = ppuControlReg1BufferBack[ppuControlReg1BufferBack.length - 1];
		for (int i = 0; i < ppuControlReg1Buffer.length; i++) {
			if (ppuControlReg1Buffer[i] >= 0) {
				ctrl = ppuControlReg1Buffer[i];
			}
			ppuControlReg1Buffer[i] = ctrl;
		}
	}
		
	private final void DrawBGToNameTblImage() {
		if (nesFile.isHorizontalMirror())
		{
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixels2(x, y, 0, 0, 0);
				}
			}
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixels2(x, y, 1, 0, 240);
				}
			}			
		} else {
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixels2(x, y, 0, 0, 0);
				}
			}
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixels2(x, y, 1, 256, 0);
				}
			}
		}
		
		System.arraycopy(scrollHBuffer, 0, scrollHBufferBack, 0, scrollHBuffer.length);
		System.arraycopy(scrollVBuffer, 0, scrollVBufferBack, 0, scrollVBuffer.length);
		System.arraycopy(ppuControlReg1Buffer, 0, ppuControlReg1BufferBack, 0, ppuControlReg1Buffer.length);
	}
	
	private final void DrawBGToNameTblImageAll() {
		if (nesFile.isHorizontalMirror())
		{
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixelsAll(x, y, 0, 0, 0);
				}
			}
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixelsAll(x, y, 1, 0, 240);
				}
			}			
		} else {
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixelsAll(x, y, 0, 0, 0);
				}
			}
			for (int y = 0; y < 30; y++) {
				for (int x = 0; x < 32; x++) {
					getVirtualPPUPixelsAll(x, y, 1, 256, 0);
				}
			}
		}
		
		System.arraycopy(scrollHBuffer, 0, scrollHBufferBack, 0, scrollHBuffer.length);
		System.arraycopy(scrollVBuffer, 0, scrollVBufferBack, 0, scrollVBuffer.length);
		System.arraycopy(ppuControlReg1Buffer, 0, ppuControlReg1BufferBack, 0, ppuControlReg1Buffer.length);
	}
		
	private final void DrawNameTblImageToScreen(Graphics g) {
		
		// wi}XN 1̏ꍇA8sNZ`悵Ȃ
		int bgOrgX = 0;
		if ((controlRegister2 & (1 << 1)) != 0) {
			bgOrgX = 1;
		}
			
		final int[] nameOffsetX = new int[] { 0, 256, 0, 256 };
		final int[] nameOffsetY = new int[] { 0, 0, 240, 240 };
				
		int y = 8;
		int oldY = 8;
		
		// \̈256x224`悷
		for (y = 8; y < 232; y++) {
			int scrollH = scrollHBuffer[y];
			int scrollV = scrollVBuffer[y];
			int nameNo = ppuControlReg1Buffer[y] & vramMirrorMask;
			
			if (scrollHBuffer[y + 1] != scrollH
					|| scrollVBuffer[y + 1] != scrollV
					|| (ppuControlReg1Buffer[y + 1] & vramMirrorMask) != nameNo)
			{
				int size = y - oldY;
				if ((nameNo & 1) == 0) {	
					g.drawImage(nameTblImage, 
							SCREEN_OFFSET_X, SCREEN_OFFSET_Y + oldY,
							nameOffsetX[nameNo] + scrollH, nameOffsetY[nameNo] + oldY,
							256, size);
				} else {
					g.drawImage(nameTblImage, 
							SCREEN_OFFSET_X, SCREEN_OFFSET_Y + oldY,
							nameOffsetX[1] + scrollH, nameOffsetY[1] + oldY,
							256 - scrollH , size);
					
					g.drawImage(nameTblImage, 
							SCREEN_OFFSET_X + 256 - scrollH, SCREEN_OFFSET_Y + oldY,
							nameOffsetX[0], nameOffsetY[0] + oldY,
							scrollH, size);
				}
				
				oldY = y;
			}
		}
		
		if (oldY < y) {	
			int scrollH = scrollHBuffer[y];
			int scrollV = scrollVBuffer[y];
			int nameNo = ppuControlReg1Buffer[y] & vramMirrorMask;
			int size = y - oldY;
			if ((nameNo & 1) == 0) {	
				g.drawImage(nameTblImage, 
						SCREEN_OFFSET_X, SCREEN_OFFSET_Y + oldY,
						nameOffsetX[nameNo] + scrollH, nameOffsetY[nameNo] + oldY,
						256, size);
			} else {
				g.drawImage(nameTblImage, 
						SCREEN_OFFSET_X, SCREEN_OFFSET_Y + oldY,
						nameOffsetX[1] + scrollH, nameOffsetY[1] + oldY,
						256 - scrollH , size);
				
				g.drawImage(nameTblImage, 
						SCREEN_OFFSET_X + 256 - scrollH, SCREEN_OFFSET_Y + oldY,
						nameOffsetX[0], nameOffsetY[0] + oldY,
						scrollH, size);
			}
		}
	}
	
	private final void getVirtualPPUPixels2(int nameX, int nameY, int nameNo, int offsetX, int offsetY)
	{
		int x = nameX << 3;
		int y = nameY << 3;
								
		// Mirroring𔽉f
		nameNo &= vramMirrorMask;
		
		// NameTableAddress OffsetvZ(1024{)
		int nameOffset = nameNo << 10;
						
		// NameTableindexvZinameTbl32x30Ȃ̂ŁA5bit shift32{j
		int nameIndex = (nameY << 5) + nameX;
		
		// AttributeTableAddress OffsetvZ
		int attOffset = nameOffset + 0x3C0;
		
		// AttTable8x7?Ȃ̂ŁA4Ŋ
		int attX = nameX >> 2;
		int attY = nameY >> 2;
		
		// AttributeTable̒lo
		int att = vram[attOffset + (attY << 3) + attX];
		
		// att̕Kv2bito
		int attShift = ((nameY & 2) << 1) + (nameX & 2);
		att = (att >> attShift) & 3;
		
		// XV̗L𒲂ׂ
		if (!isDirty[nameOffset + nameIndex] && !isDirtyPaletteSet[att]) {			
			return;
		}
		isDirty[nameOffset + nameIndex] = false;
		
		// NameTablecharNoo
		int charNo = vram[nameOffset + nameIndex];
				
		// Rg[WX^1Bit4Awip^[e[uAhXw擾
		// 0:$0000A1:$1000
//		int patternAdr = (controlRegister1 & (1 << 4)) << 8;
		int patternAdr = (ppuControlReg1Buffer[y]  & (1 << 4)) << 8;	
		
		// ʉߐFx^̃LN^ꍇ
		if (isTransparentOnly[(patternAdr >> 4) + charNo]) {
			nameTblGraphics.setColor(paletteData[palette[16]]);
			nameTblGraphics.fillRect(offsetX + x, offsetY + y, 8, 8);
			//g.setColor(paletteData[palette[16]]);
			//g.fillRect(offsetX + x, offsetY + y, 8, 8);
			return;
		}
						
		// att̒lŁAparetteTable擾
		pal2rgb = pal2rgbTbl[att];
		
		// chrRom32Adderss offsetvZ
		bgChrRom32Index = (patternAdr + (charNo << 4)) >> 2;
				
		DrawOneCharFlipNone();
		nameTblGraphics.setPixels(offsetX + x, offsetY + y, 8, 8, buffer, 0);
//		g.setPixels(offsetX + x, offsetY + y, 8, 8, buffer, 0);		
	}

	private final void getVirtualPPUPixelsAll(int nameX, int nameY, int nameNo, int offsetX, int offsetY)
	{
		int x = nameX << 3;
		int y = nameY << 3;
								
		// Mirroring𔽉f
		nameNo &= vramMirrorMask;
		
		// NameTableAddress OffsetvZ(1024{)
		int nameOffset = nameNo << 10;
						
		// NameTableindexvZinameTbl32x30Ȃ̂ŁA5bit shift32{j
		int nameIndex = (nameY << 5) + nameX;
		
		// AttributeTableAddress OffsetvZ
		int attOffset = nameOffset + 0x3C0;
		
		// AttTable8x7?Ȃ̂ŁA4Ŋ
		int attX = nameX >> 2;
		int attY = nameY >> 2;
		
		// AttributeTable̒lo
		int att = vram[attOffset + (attY << 3) + attX];
		
		// att̕Kv2bito
		int attShift = ((nameY & 2) << 1) + (nameX & 2);
		att = (att >> attShift) & 3;
		
		isDirty[nameOffset + nameIndex] = false;
		
		// NameTablecharNoo
		int charNo = vram[nameOffset + nameIndex];
				
		// Rg[WX^1Bit4Awip^[e[uAhXw擾
		// 0:$0000A1:$1000
//		int patternAdr = (controlRegister1 & (1 << 4)) << 8;
		int patternAdr = (ppuControlReg1Buffer[y]  & (1 << 4)) << 8;	
		
		// ʉߐFx^̃LN^ꍇ
		if (isTransparentOnly[(patternAdr >> 4) + charNo]) {
			nameTblGraphics.setColor(paletteData[palette[16]]);
			nameTblGraphics.fillRect(offsetX + x, offsetY + y, 8, 8);
			//g.setColor(paletteData[palette[16]]);
			//g.fillRect(offsetX + x, offsetY + y, 8, 8);
			return;
		}
						
		// att̒lŁAparetteTable擾
		pal2rgb = pal2rgbTbl[att];
		
		// chrRom32Adderss offsetvZ
		bgChrRom32Index = (patternAdr + (charNo << 4)) >> 2;
				
		DrawOneCharFlipNone();
		nameTblGraphics.setPixels(offsetX + x, offsetY + y, 8, 8, buffer, 0);
//		g.setPixels(offsetX + x, offsetY + y, 8, 8, buffer, 0);		
	}

	/**
	 * XvCg`悵܂B
	 */
	private final void DrawSprite8(Graphics g) {		

		// XvCg}XN 1̏ꍇAʍ8sNZ`悵Ȃ
		int spriteMaskX = -1;
		if ((controlRegister2 & (1 << 2)) != 0) {
			spriteMaskX = 0;
		}

		// XvCg̕`
		for (int i = 0; i < 64; i++) {
			final int adr = i << 2;
			final int sprY = spriteRam[adr] + 1;
			final int charNo = spriteRam[adr + 1];
			final int att = spriteRam[adr + 2];
			final int sprX = spriteRam[adr + 3];
			
			if (sprX == spriteMaskX) {
				continue;
			}
			
			// bit3 XvCgp^[e[uAhXw 0:$0000A1:$1000
			int patternAdr = 0;
			if ((ppuControlReg1Buffer[sprY] & (1 << 3)) != 0) {
				patternAdr = 0x1000 >> 2;
			}
			
			if (isTransparentOnly[(patternAdr >> 2) + charNo]) {
				continue;
			}
						
			if ((charNo != spriteRamBack[adr + 1]
				|| att != spriteRamBack[adr + 2])
				&& (att & (1 << 5)) == 0) {
				pal2rgb = pal2rgbTbl[(att & 3) + 4];
				bgChrRom32Index = patternAdr + (charNo << 2);

				final int flip = att >>> 6;
				if (flip == 0) {
					DrawOneCharFlipNone();						
				} else if (flip == 1) {
					DrawOneCharFlipHorizontal();
				} else if (flip == 2) {
					DrawOneCharFlipVertical();
				} else if (flip == 3) {
					DrawOneCharFlipRotate();
				}
				spriteGraphics[i].setPixels(0, 0, 8, 8, buffer, 0);
				spriteImg[i].setTransparentColor(paletteData[palette[16]]);
				spriteImg[i].setTransparentEnabled(true);				
			}
						
			final int drawX = sprX + SCREEN_OFFSET_X;
			final int drawY = sprY + SCREEN_OFFSET_Y;
			if ((att & (1 << 5)) == 0) {
				// SpriteD				
				g.drawImage(spriteImg[i], drawX, drawY, 0, 0, 8, 8);
			} else {
				// BGD
				pal2rgb = pal2rgbTbl[(att & 3) + 4];
				bgChrRom32Index = patternAdr + (charNo << 2);				
				g.getPixels(drawX, drawY, 8, 8, buffer, 0);
				
				final int flip = att >>> 6;
				if (flip == 0) {
					drawBGBackCharFlipNone();						
				} else if (flip == 1) {
					drawBGBackCharFlipHorizontal();
				} else if (flip == 2) {
					drawBGBackCharFlipVertical();
				} else if (flip == 3) {
					drawBGBackCharFlipRotate();
				}
				
				g.setPixels(drawX, drawY, 8, 8, buffer, 0);
			}
		}
	}

	private final void DrawSprite16(Graphics g) {		
		// bit3 XvCgp^[e[uAhXw 0:$0000A1:$1000
		int patternAdr0;
		int patternAdr1;		
		
		// XvCg}XN 1̏ꍇAʍ8sNZ`悵Ȃ
		int spriteMaskX = -1;
		if ((controlRegister2 & (1 << 2)) != 0) {
			spriteMaskX = 0;
		}

		// XvCg̕`
		for (int i = 0; i < 64; i++) {
			final int adr = i << 2;
			final int sprY = spriteRam[adr] + 1;
			final int charNo = spriteRam[adr + 1];
			final int att = spriteRam[adr + 2];
			final int sprX = spriteRam[adr + 3];
			
			if (sprX == spriteMaskX) {
				continue;
			}
			
			if ((ppuControlReg1Buffer[sprY] & (1 << 3)) != 0) {
				patternAdr0 = 0x1000 >> 2;
				patternAdr1 = 0;
			} else {
				patternAdr0 = 0;
				patternAdr1 = 0x1000 >> 2;			
			}

			
			if (isTransparentOnly[(patternAdr0 >> 2) + charNo]
			     && isTransparentOnly[(patternAdr1 >> 2) + charNo]) {
				continue;
			}
						
			if ((charNo != spriteRamBack[adr + 1]
				|| att != spriteRamBack[adr + 2])
				&& (att & (1 << 5)) == 0) {
				
				pal2rgb = pal2rgbTbl[(att & 3) + 4];
				bgChrRom32Index = patternAdr0 + (charNo << 2);
				final int flip = att >>> 6;
				if (flip == 0) {
					DrawOneCharFlipNone();						
				} else if (flip == 1) {
					DrawOneCharFlipHorizontal();
				} else if (flip == 2) {
					DrawOneCharFlipVertical();
				} else if (flip == 3) {
					DrawOneCharFlipRotate();
				}
				spriteGraphics[i].setPixels(0, 0, 8, 8, buffer, 0);
				
				bgChrRom32Index = patternAdr1 + (charNo << 2);
				if (flip == 0) {
					DrawOneCharFlipNone();						
				} else if (flip == 1) {
					DrawOneCharFlipHorizontal();
				} else if (flip == 2) {
					DrawOneCharFlipVertical();
				} else if (flip == 3) {
					DrawOneCharFlipRotate();
				}
				spriteGraphics[i].setPixels(0, 8, 8, 8, buffer, 0);

				spriteImg[i].setTransparentColor(paletteData[palette[16]]);
				spriteImg[i].setTransparentEnabled(true);				
			}
						
			final int drawX = sprX + SCREEN_OFFSET_X;
			final int drawY = sprY + SCREEN_OFFSET_Y;
			if ((att & (1 << 5)) == 0) {
				// SpriteD				
				g.drawImage(spriteImg[i], drawX, drawY, 0, 0, 8, 16);
			} else {
				// BGD
				pal2rgb = pal2rgbTbl[(att & 3) + 4];
				final int flip = att >>> 6;
				
				bgChrRom32Index = patternAdr0 + (charNo << 2);
				g.getPixels(drawX, drawY, 8, 8, buffer, 0);				
				if (flip == 0) {
					drawBGBackCharFlipNone();						
				} else if (flip == 1) {
					drawBGBackCharFlipHorizontal();
				} else if (flip == 2) {
					drawBGBackCharFlipVertical();
				} else if (flip == 3) {
					drawBGBackCharFlipRotate();
				}
				g.setPixels(drawX, drawY, 8, 8, buffer, 0);

				bgChrRom32Index = patternAdr1 + (charNo << 2);
				g.getPixels(drawX, drawY + 8, 8, 8, buffer, 0);				
				if (flip == 0) {
					drawBGBackCharFlipNone();						
				} else if (flip == 1) {
					drawBGBackCharFlipHorizontal();
				} else if (flip == 2) {
					drawBGBackCharFlipVertical();
				} else if (flip == 3) {
					drawBGBackCharFlipRotate();
				}
				g.setPixels(drawX, drawY + 8, 8, 8, buffer, 0);
			}
		}
	}

	private final void DrawOneCharFlipNone() {
		int pat;
		final int buffer[] = this.buffer;		
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		buffer[ 0] = pal2rgb[(pat = chrRom32[bgChrRom32Index]) & 3];
		buffer[ 1] = pal2rgb[(pat >>>  2) & 3];
		buffer[ 2] = pal2rgb[(pat >>>  4) & 3];
		buffer[ 3] = pal2rgb[(pat >>>  6) & 3];
		buffer[ 4] = pal2rgb[(pat >>>  8) & 3];
		buffer[ 5] = pal2rgb[(pat >>> 10) & 3];
		buffer[ 6] = pal2rgb[(pat >>> 12) & 3];
		buffer[ 7] = pal2rgb[(pat >>> 14) & 3];
		buffer[ 8] = pal2rgb[(pat >>> 16) & 3];
		buffer[ 9] = pal2rgb[(pat >>> 18) & 3];
		buffer[10] = pal2rgb[(pat >>> 20) & 3];
		buffer[11] = pal2rgb[(pat >>> 22) & 3];
		buffer[12] = pal2rgb[(pat >>> 24) & 3];
		buffer[13] = pal2rgb[(pat >>> 26) & 3];
		buffer[14] = pal2rgb[(pat >>> 28) & 3];
		buffer[15] = pal2rgb[(pat >>> 30)    ];
		buffer[16] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 1]) & 3];
		buffer[17] = pal2rgb[(pat >>>  2) & 3];
		buffer[18] = pal2rgb[(pat >>>  4) & 3];
		buffer[19] = pal2rgb[(pat >>>  6) & 3];
		buffer[20] = pal2rgb[(pat >>>  8) & 3];
		buffer[21] = pal2rgb[(pat >>> 10) & 3];
		buffer[22] = pal2rgb[(pat >>> 12) & 3];
		buffer[23] = pal2rgb[(pat >>> 14) & 3];
		buffer[24] = pal2rgb[(pat >>> 16) & 3];
		buffer[25] = pal2rgb[(pat >>> 18) & 3];
		buffer[26] = pal2rgb[(pat >>> 20) & 3];
		buffer[27] = pal2rgb[(pat >>> 22) & 3];
		buffer[28] = pal2rgb[(pat >>> 24) & 3];
		buffer[29] = pal2rgb[(pat >>> 26) & 3];
		buffer[30] = pal2rgb[(pat >>> 28) & 3];
		buffer[31] = pal2rgb[(pat >>> 30)    ];
		buffer[32] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 2]) & 3];
		buffer[33] = pal2rgb[(pat >>>  2) & 3];
		buffer[34] = pal2rgb[(pat >>>  4) & 3];
		buffer[35] = pal2rgb[(pat >>>  6) & 3];
		buffer[36] = pal2rgb[(pat >>>  8) & 3];
		buffer[37] = pal2rgb[(pat >>> 10) & 3];
		buffer[38] = pal2rgb[(pat >>> 12) & 3];
		buffer[39] = pal2rgb[(pat >>> 14) & 3];
		buffer[40] = pal2rgb[(pat >>> 16) & 3];
		buffer[41] = pal2rgb[(pat >>> 18) & 3];
		buffer[42] = pal2rgb[(pat >>> 20) & 3];
		buffer[43] = pal2rgb[(pat >>> 22) & 3];
		buffer[44] = pal2rgb[(pat >>> 24) & 3];
		buffer[45] = pal2rgb[(pat >>> 26) & 3];
		buffer[46] = pal2rgb[(pat >>> 28) & 3];
		buffer[47] = pal2rgb[(pat >>> 30)    ];
		buffer[48] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 3]) & 3];
		buffer[49] = pal2rgb[(pat >>>  2) & 3];
		buffer[50] = pal2rgb[(pat >>>  4) & 3];
		buffer[51] = pal2rgb[(pat >>>  6) & 3];
		buffer[52] = pal2rgb[(pat >>>  8) & 3];
		buffer[53] = pal2rgb[(pat >>> 10) & 3];
		buffer[54] = pal2rgb[(pat >>> 12) & 3];
		buffer[55] = pal2rgb[(pat >>> 14) & 3];
		buffer[56] = pal2rgb[(pat >>> 16) & 3];
		buffer[57] = pal2rgb[(pat >>> 18) & 3];
		buffer[58] = pal2rgb[(pat >>> 20) & 3];
		buffer[59] = pal2rgb[(pat >>> 22) & 3];
		buffer[60] = pal2rgb[(pat >>> 24) & 3];
		buffer[61] = pal2rgb[(pat >>> 26) & 3];
		buffer[62] = pal2rgb[(pat >>> 28) & 3];
		buffer[63] = pal2rgb[(pat >>> 30)    ];
	}
	
	private final void DrawOneCharFlipHorizontal() {
		int pat;
		final int buffer[] = this.buffer;		
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		buffer[ 7] = pal2rgb[(pat = chrRom32[bgChrRom32Index]) & 3];
		buffer[ 6] = pal2rgb[(pat >>>  2) & 3];
		buffer[ 5] = pal2rgb[(pat >>>  4) & 3];
		buffer[ 4] = pal2rgb[(pat >>>  6) & 3];
		buffer[ 3] = pal2rgb[(pat >>>  8) & 3];
		buffer[ 2] = pal2rgb[(pat >>> 10) & 3];
		buffer[ 1] = pal2rgb[(pat >>> 12) & 3];
		buffer[ 0] = pal2rgb[(pat >>> 14) & 3];
		buffer[15] = pal2rgb[(pat >>> 16) & 3];
		buffer[14] = pal2rgb[(pat >>> 18) & 3];
		buffer[13] = pal2rgb[(pat >>> 20) & 3];
		buffer[12] = pal2rgb[(pat >>> 22) & 3];
		buffer[11] = pal2rgb[(pat >>> 24) & 3];
		buffer[10] = pal2rgb[(pat >>> 26) & 3];
		buffer[ 9] = pal2rgb[(pat >>> 28) & 3];
		buffer[ 8] = pal2rgb[(pat >>> 30)    ];
		buffer[23] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 1]) & 3];
		buffer[22] = pal2rgb[(pat >>>  2) & 3];
		buffer[21] = pal2rgb[(pat >>>  4) & 3];
		buffer[20] = pal2rgb[(pat >>>  6) & 3];
		buffer[19] = pal2rgb[(pat >>>  8) & 3];
		buffer[18] = pal2rgb[(pat >>> 10) & 3];
		buffer[17] = pal2rgb[(pat >>> 12) & 3];
		buffer[16] = pal2rgb[(pat >>> 14) & 3];
		buffer[31] = pal2rgb[(pat >>> 16) & 3];
		buffer[30] = pal2rgb[(pat >>> 18) & 3];
		buffer[29] = pal2rgb[(pat >>> 20) & 3];
		buffer[28] = pal2rgb[(pat >>> 22) & 3];
		buffer[27] = pal2rgb[(pat >>> 24) & 3];
		buffer[26] = pal2rgb[(pat >>> 26) & 3];
		buffer[25] = pal2rgb[(pat >>> 28) & 3];
		buffer[24] = pal2rgb[(pat >>> 30)    ];
		buffer[39] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 2]) & 3];
		buffer[38] = pal2rgb[(pat >>>  2) & 3];
		buffer[37] = pal2rgb[(pat >>>  4) & 3];
		buffer[36] = pal2rgb[(pat >>>  6) & 3];
		buffer[35] = pal2rgb[(pat >>>  8) & 3];
		buffer[34] = pal2rgb[(pat >>> 10) & 3];
		buffer[33] = pal2rgb[(pat >>> 12) & 3];
		buffer[32] = pal2rgb[(pat >>> 14) & 3];
		buffer[47] = pal2rgb[(pat >>> 16) & 3];
		buffer[46] = pal2rgb[(pat >>> 18) & 3];
		buffer[45] = pal2rgb[(pat >>> 20) & 3];
		buffer[44] = pal2rgb[(pat >>> 22) & 3];
		buffer[43] = pal2rgb[(pat >>> 24) & 3];
		buffer[42] = pal2rgb[(pat >>> 26) & 3];
		buffer[41] = pal2rgb[(pat >>> 28) & 3];
		buffer[40] = pal2rgb[(pat >>> 30)    ];
		buffer[55] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 3]) & 3];
		buffer[54] = pal2rgb[(pat >>>  2) & 3];
		buffer[53] = pal2rgb[(pat >>>  4) & 3];
		buffer[52] = pal2rgb[(pat >>>  6) & 3];
		buffer[51] = pal2rgb[(pat >>>  8) & 3];
		buffer[50] = pal2rgb[(pat >>> 10) & 3];
		buffer[49] = pal2rgb[(pat >>> 12) & 3];
		buffer[48] = pal2rgb[(pat >>> 14) & 3];
		buffer[63] = pal2rgb[(pat >>> 16) & 3];
		buffer[62] = pal2rgb[(pat >>> 18) & 3];
		buffer[61] = pal2rgb[(pat >>> 20) & 3];
		buffer[60] = pal2rgb[(pat >>> 22) & 3];
		buffer[59] = pal2rgb[(pat >>> 24) & 3];
		buffer[58] = pal2rgb[(pat >>> 26) & 3];
		buffer[57] = pal2rgb[(pat >>> 28) & 3];
		buffer[56] = pal2rgb[(pat >>> 30)    ];
	}
	
	private final void DrawOneCharFlipVertical() {
		int pat;
		final int buffer[] = this.buffer;		
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		buffer[56] = pal2rgb[(pat = chrRom32[bgChrRom32Index]) & 3];
		buffer[57] = pal2rgb[(pat >>>  2) & 3];
		buffer[58] = pal2rgb[(pat >>>  4) & 3];
		buffer[59] = pal2rgb[(pat >>>  6) & 3];
		buffer[60] = pal2rgb[(pat >>>  8) & 3];
		buffer[61] = pal2rgb[(pat >>> 10) & 3];
		buffer[62] = pal2rgb[(pat >>> 12) & 3];
		buffer[63] = pal2rgb[(pat >>> 14) & 3];
		buffer[48] = pal2rgb[(pat >>> 16) & 3];
		buffer[49] = pal2rgb[(pat >>> 18) & 3];
		buffer[50] = pal2rgb[(pat >>> 20) & 3];
		buffer[51] = pal2rgb[(pat >>> 22) & 3];
		buffer[52] = pal2rgb[(pat >>> 24) & 3];
		buffer[53] = pal2rgb[(pat >>> 26) & 3];
		buffer[54] = pal2rgb[(pat >>> 28) & 3];
		buffer[55] = pal2rgb[(pat >>> 30)    ];
		buffer[40] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 1]) & 3];
		buffer[41] = pal2rgb[(pat >>>  2) & 3];
		buffer[42] = pal2rgb[(pat >>>  4) & 3];
		buffer[43] = pal2rgb[(pat >>>  6) & 3];
		buffer[44] = pal2rgb[(pat >>>  8) & 3];
		buffer[45] = pal2rgb[(pat >>> 10) & 3];
		buffer[46] = pal2rgb[(pat >>> 12) & 3];
		buffer[47] = pal2rgb[(pat >>> 14) & 3];
		buffer[32] = pal2rgb[(pat >>> 16) & 3];
		buffer[33] = pal2rgb[(pat >>> 18) & 3];
		buffer[34] = pal2rgb[(pat >>> 20) & 3];
		buffer[35] = pal2rgb[(pat >>> 22) & 3];
		buffer[36] = pal2rgb[(pat >>> 24) & 3];
		buffer[37] = pal2rgb[(pat >>> 26) & 3];
		buffer[38] = pal2rgb[(pat >>> 28) & 3];
		buffer[39] = pal2rgb[(pat >>> 30)    ];
		buffer[24] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 2]) & 3];
		buffer[25] = pal2rgb[(pat >>>  2) & 3];
		buffer[26] = pal2rgb[(pat >>>  4) & 3];
		buffer[27] = pal2rgb[(pat >>>  6) & 3];
		buffer[28] = pal2rgb[(pat >>>  8) & 3];
		buffer[29] = pal2rgb[(pat >>> 10) & 3];
		buffer[30] = pal2rgb[(pat >>> 12) & 3];
		buffer[31] = pal2rgb[(pat >>> 14) & 3];
		buffer[16] = pal2rgb[(pat >>> 16) & 3];
		buffer[17] = pal2rgb[(pat >>> 18) & 3];
		buffer[18] = pal2rgb[(pat >>> 20) & 3];
		buffer[19] = pal2rgb[(pat >>> 22) & 3];
		buffer[20] = pal2rgb[(pat >>> 24) & 3];
		buffer[21] = pal2rgb[(pat >>> 26) & 3];
		buffer[22] = pal2rgb[(pat >>> 28) & 3];
		buffer[23] = pal2rgb[(pat >>> 30)    ];
		buffer[ 8] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 3]) & 3];
		buffer[ 9] = pal2rgb[(pat >>>  2) & 3];
		buffer[10] = pal2rgb[(pat >>>  4) & 3];
		buffer[11] = pal2rgb[(pat >>>  6) & 3];
		buffer[12] = pal2rgb[(pat >>>  8) & 3];
		buffer[13] = pal2rgb[(pat >>> 10) & 3];
		buffer[14] = pal2rgb[(pat >>> 12) & 3];
		buffer[15] = pal2rgb[(pat >>> 14) & 3];
		buffer[ 0] = pal2rgb[(pat >>> 16) & 3];
		buffer[ 1] = pal2rgb[(pat >>> 18) & 3];
		buffer[ 2] = pal2rgb[(pat >>> 20) & 3];
		buffer[ 3] = pal2rgb[(pat >>> 22) & 3];
		buffer[ 4] = pal2rgb[(pat >>> 24) & 3];
		buffer[ 5] = pal2rgb[(pat >>> 26) & 3];
		buffer[ 6] = pal2rgb[(pat >>> 28) & 3];
		buffer[ 7] = pal2rgb[(pat >>> 30)    ];
	}
	
	private final void DrawOneCharFlipRotate() {
		int pat;
		final int buffer[] = this.buffer;		
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		buffer[63] = pal2rgb[(pat = chrRom32[bgChrRom32Index]) & 3];
		buffer[62] = pal2rgb[(pat >>>  2) & 3];
		buffer[61] = pal2rgb[(pat >>>  4) & 3];
		buffer[60] = pal2rgb[(pat >>>  6) & 3];
		buffer[59] = pal2rgb[(pat >>>  8) & 3];
		buffer[58] = pal2rgb[(pat >>> 10) & 3];
		buffer[57] = pal2rgb[(pat >>> 12) & 3];
		buffer[56] = pal2rgb[(pat >>> 14) & 3];
		buffer[55] = pal2rgb[(pat >>> 16) & 3];
		buffer[54] = pal2rgb[(pat >>> 18) & 3];
		buffer[53] = pal2rgb[(pat >>> 20) & 3];
		buffer[52] = pal2rgb[(pat >>> 22) & 3];
		buffer[51] = pal2rgb[(pat >>> 24) & 3];
		buffer[50] = pal2rgb[(pat >>> 26) & 3];
		buffer[49] = pal2rgb[(pat >>> 28) & 3];
		buffer[48] = pal2rgb[(pat >>> 30)    ];
		buffer[47] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 1]) & 3];
		buffer[46] = pal2rgb[(pat >>>  2) & 3];
		buffer[45] = pal2rgb[(pat >>>  4) & 3];
		buffer[44] = pal2rgb[(pat >>>  6) & 3];
		buffer[43] = pal2rgb[(pat >>>  8) & 3];
		buffer[42] = pal2rgb[(pat >>> 10) & 3];
		buffer[41] = pal2rgb[(pat >>> 12) & 3];
		buffer[40] = pal2rgb[(pat >>> 14) & 3];
		buffer[39] = pal2rgb[(pat >>> 16) & 3];
		buffer[38] = pal2rgb[(pat >>> 18) & 3];
		buffer[37] = pal2rgb[(pat >>> 20) & 3];
		buffer[36] = pal2rgb[(pat >>> 22) & 3];
		buffer[35] = pal2rgb[(pat >>> 24) & 3];
		buffer[34] = pal2rgb[(pat >>> 26) & 3];
		buffer[33] = pal2rgb[(pat >>> 28) & 3];
		buffer[32] = pal2rgb[(pat >>> 30)    ];
		buffer[31] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 2]) & 3];
		buffer[30] = pal2rgb[(pat >>>  2) & 3];
		buffer[29] = pal2rgb[(pat >>>  4) & 3];
		buffer[28] = pal2rgb[(pat >>>  6) & 3];
		buffer[27] = pal2rgb[(pat >>>  8) & 3];
		buffer[26] = pal2rgb[(pat >>> 10) & 3];
		buffer[25] = pal2rgb[(pat >>> 12) & 3];
		buffer[24] = pal2rgb[(pat >>> 14) & 3];
		buffer[23] = pal2rgb[(pat >>> 16) & 3];
		buffer[22] = pal2rgb[(pat >>> 18) & 3];
		buffer[21] = pal2rgb[(pat >>> 20) & 3];
		buffer[20] = pal2rgb[(pat >>> 22) & 3];
		buffer[19] = pal2rgb[(pat >>> 24) & 3];
		buffer[18] = pal2rgb[(pat >>> 26) & 3];
		buffer[17] = pal2rgb[(pat >>> 28) & 3];
		buffer[16] = pal2rgb[(pat >>> 30)    ];
		buffer[15] = pal2rgb[(pat = chrRom32[bgChrRom32Index + 3]) & 3];
		buffer[14] = pal2rgb[(pat >>>  2) & 3];
		buffer[13] = pal2rgb[(pat >>>  4) & 3];
		buffer[12] = pal2rgb[(pat >>>  6) & 3];
		buffer[11] = pal2rgb[(pat >>>  8) & 3];
		buffer[10] = pal2rgb[(pat >>> 10) & 3];
		buffer[9 ] = pal2rgb[(pat >>> 12) & 3];
		buffer[8 ] = pal2rgb[(pat >>> 14) & 3];
		buffer[7 ] = pal2rgb[(pat >>> 16) & 3];
		buffer[6 ] = pal2rgb[(pat >>> 18) & 3];
		buffer[5 ] = pal2rgb[(pat >>> 20) & 3];
		buffer[4 ] = pal2rgb[(pat >>> 22) & 3];
		buffer[3 ] = pal2rgb[(pat >>> 24) & 3];
		buffer[2 ] = pal2rgb[(pat >>> 26) & 3];
		buffer[1 ] = pal2rgb[(pat >>> 28) & 3];
		buffer[0 ] = pal2rgb[(pat >>> 30)    ];
	}
	
	private final void drawBGBackCharFlipNone() {
		int pat;
		final int buffer[] = this.buffer;
		final int transparentColor = paletteData[palette[16]];
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		pat = chrRom32[bgChrRom32Index];
		if (buffer[0 ] == transparentColor) buffer[0 ] = pal2rgb[(pat       ) & 3];
		if (buffer[1 ] == transparentColor) buffer[1 ] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[2 ] == transparentColor) buffer[2 ] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[3 ] == transparentColor) buffer[3 ] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[4 ] == transparentColor) buffer[4 ] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[5 ] == transparentColor) buffer[5 ] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[6 ] == transparentColor) buffer[6 ] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[7 ] == transparentColor) buffer[7 ] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[8 ] == transparentColor) buffer[8 ] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[9 ] == transparentColor) buffer[9 ] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[10] == transparentColor) buffer[10] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[11] == transparentColor) buffer[11] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[12] == transparentColor) buffer[12] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[13] == transparentColor) buffer[13] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[14] == transparentColor) buffer[14] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[15] == transparentColor) buffer[15] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 1];
		if (buffer[16] == transparentColor) buffer[16] = pal2rgb[(pat       ) & 3];
		if (buffer[17] == transparentColor) buffer[17] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[18] == transparentColor) buffer[18] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[19] == transparentColor) buffer[19] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[20] == transparentColor) buffer[20] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[21] == transparentColor) buffer[21] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[22] == transparentColor) buffer[22] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[23] == transparentColor) buffer[23] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[24] == transparentColor) buffer[24] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[25] == transparentColor) buffer[25] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[26] == transparentColor) buffer[26] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[27] == transparentColor) buffer[27] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[28] == transparentColor) buffer[28] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[29] == transparentColor) buffer[29] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[30] == transparentColor) buffer[30] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[31] == transparentColor) buffer[31] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 2];
		if (buffer[32] == transparentColor) buffer[32] = pal2rgb[(pat       ) & 3];
		if (buffer[33] == transparentColor) buffer[33] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[34] == transparentColor) buffer[34] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[35] == transparentColor) buffer[35] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[36] == transparentColor) buffer[36] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[37] == transparentColor) buffer[37] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[38] == transparentColor) buffer[38] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[39] == transparentColor) buffer[39] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[40] == transparentColor) buffer[40] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[41] == transparentColor) buffer[41] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[42] == transparentColor) buffer[42] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[43] == transparentColor) buffer[43] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[44] == transparentColor) buffer[44] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[45] == transparentColor) buffer[45] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[46] == transparentColor) buffer[46] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[47] == transparentColor) buffer[47] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 3];
		if (buffer[48] == transparentColor) buffer[48] = pal2rgb[(pat       ) & 3];
		if (buffer[49] == transparentColor) buffer[49] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[50] == transparentColor) buffer[50] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[51] == transparentColor) buffer[51] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[52] == transparentColor) buffer[52] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[53] == transparentColor) buffer[53] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[54] == transparentColor) buffer[54] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[55] == transparentColor) buffer[55] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[56] == transparentColor) buffer[56] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[57] == transparentColor) buffer[57] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[58] == transparentColor) buffer[58] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[59] == transparentColor) buffer[59] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[60] == transparentColor) buffer[60] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[61] == transparentColor) buffer[61] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[62] == transparentColor) buffer[62] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[63] == transparentColor) buffer[63] = pal2rgb[(pat >>> 30)    ];
	}

	private final void drawBGBackCharFlipHorizontal() {
		int pat;
		final int buffer[] = this.buffer;
		final int transparentColor = paletteData[palette[16]];
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		pat = chrRom32[bgChrRom32Index];
		if (buffer[ 7] == transparentColor) buffer[ 7] = pal2rgb[(pat       ) & 3];
		if (buffer[ 6] == transparentColor) buffer[ 6] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[ 5] == transparentColor) buffer[ 5] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[ 4] == transparentColor) buffer[ 4] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[ 3] == transparentColor) buffer[ 3] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[ 2] == transparentColor) buffer[ 2] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[ 1] == transparentColor) buffer[ 1] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[ 0] == transparentColor) buffer[ 0] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[15] == transparentColor) buffer[15] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[14] == transparentColor) buffer[14] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[13] == transparentColor) buffer[13] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[12] == transparentColor) buffer[12] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[11] == transparentColor) buffer[11] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[10] == transparentColor) buffer[10] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[ 9] == transparentColor) buffer[ 9] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[ 8] == transparentColor) buffer[ 8] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 1];
		if (buffer[23] == transparentColor) buffer[23] = pal2rgb[(pat       ) & 3];
		if (buffer[22] == transparentColor) buffer[22] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[21] == transparentColor) buffer[21] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[20] == transparentColor) buffer[20] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[19] == transparentColor) buffer[19] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[18] == transparentColor) buffer[18] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[17] == transparentColor) buffer[17] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[16] == transparentColor) buffer[16] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[31] == transparentColor) buffer[31] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[30] == transparentColor) buffer[30] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[29] == transparentColor) buffer[29] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[28] == transparentColor) buffer[28] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[27] == transparentColor) buffer[27] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[26] == transparentColor) buffer[26] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[25] == transparentColor) buffer[25] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[24] == transparentColor) buffer[24] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 2];
		if (buffer[39] == transparentColor) buffer[39] = pal2rgb[(pat       ) & 3];
		if (buffer[38] == transparentColor) buffer[38] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[37] == transparentColor) buffer[37] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[36] == transparentColor) buffer[36] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[35] == transparentColor) buffer[35] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[34] == transparentColor) buffer[34] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[33] == transparentColor) buffer[33] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[32] == transparentColor) buffer[32] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[47] == transparentColor) buffer[47] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[46] == transparentColor) buffer[46] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[45] == transparentColor) buffer[45] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[44] == transparentColor) buffer[44] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[43] == transparentColor) buffer[43] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[42] == transparentColor) buffer[42] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[41] == transparentColor) buffer[41] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[40] == transparentColor) buffer[40] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 3];
		if (buffer[55] == transparentColor) buffer[55] = pal2rgb[(pat       ) & 3];
		if (buffer[54] == transparentColor) buffer[54] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[53] == transparentColor) buffer[53] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[52] == transparentColor) buffer[52] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[51] == transparentColor) buffer[51] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[50] == transparentColor) buffer[50] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[49] == transparentColor) buffer[49] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[48] == transparentColor) buffer[48] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[63] == transparentColor) buffer[63] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[62] == transparentColor) buffer[62] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[61] == transparentColor) buffer[61] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[60] == transparentColor) buffer[60] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[59] == transparentColor) buffer[59] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[58] == transparentColor) buffer[58] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[57] == transparentColor) buffer[57] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[56] == transparentColor) buffer[56] = pal2rgb[(pat >>> 30)    ];

	}
	
	private final void drawBGBackCharFlipVertical() {
		int pat;
		final int buffer[] = this.buffer;
		final int transparentColor = paletteData[palette[16]];
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		pat = chrRom32[bgChrRom32Index];
		if (buffer[56] == transparentColor) buffer[56] = pal2rgb[(pat       ) & 3];
		if (buffer[57] == transparentColor) buffer[57] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[58] == transparentColor) buffer[58] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[59] == transparentColor) buffer[59] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[60] == transparentColor) buffer[60] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[61] == transparentColor) buffer[61] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[62] == transparentColor) buffer[62] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[63] == transparentColor) buffer[63] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[48] == transparentColor) buffer[48] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[49] == transparentColor) buffer[49] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[50] == transparentColor) buffer[50] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[51] == transparentColor) buffer[51] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[52] == transparentColor) buffer[52] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[53] == transparentColor) buffer[53] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[54] == transparentColor) buffer[54] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[55] == transparentColor) buffer[55] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 1];
		if (buffer[40] == transparentColor) buffer[40] = pal2rgb[(pat       ) & 3];
		if (buffer[41] == transparentColor) buffer[41] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[42] == transparentColor) buffer[42] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[43] == transparentColor) buffer[43] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[44] == transparentColor) buffer[44] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[45] == transparentColor) buffer[45] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[46] == transparentColor) buffer[46] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[47] == transparentColor) buffer[47] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[32] == transparentColor) buffer[32] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[33] == transparentColor) buffer[33] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[34] == transparentColor) buffer[34] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[35] == transparentColor) buffer[35] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[36] == transparentColor) buffer[36] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[37] == transparentColor) buffer[37] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[38] == transparentColor) buffer[38] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[39] == transparentColor) buffer[39] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 2];
		if (buffer[24] == transparentColor) buffer[24] = pal2rgb[(pat       ) & 3];
		if (buffer[25] == transparentColor) buffer[25] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[26] == transparentColor) buffer[26] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[27] == transparentColor) buffer[27] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[28] == transparentColor) buffer[28] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[29] == transparentColor) buffer[29] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[30] == transparentColor) buffer[30] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[31] == transparentColor) buffer[31] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[16] == transparentColor) buffer[16] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[17] == transparentColor) buffer[17] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[18] == transparentColor) buffer[18] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[19] == transparentColor) buffer[19] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[20] == transparentColor) buffer[20] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[21] == transparentColor) buffer[21] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[22] == transparentColor) buffer[22] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[23] == transparentColor) buffer[23] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 3];
		if (buffer[ 8] == transparentColor) buffer[ 8] = pal2rgb[(pat       ) & 3];
		if (buffer[ 9] == transparentColor) buffer[ 9] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[10] == transparentColor) buffer[10] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[11] == transparentColor) buffer[11] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[12] == transparentColor) buffer[12] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[13] == transparentColor) buffer[13] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[14] == transparentColor) buffer[14] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[15] == transparentColor) buffer[15] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[ 0] == transparentColor) buffer[ 0] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[ 1] == transparentColor) buffer[ 1] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[ 2] == transparentColor) buffer[ 2] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[ 3] == transparentColor) buffer[ 3] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[ 4] == transparentColor) buffer[ 4] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[ 5] == transparentColor) buffer[ 5] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[ 6] == transparentColor) buffer[ 6] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[ 7] == transparentColor) buffer[ 7] = pal2rgb[(pat >>> 30)    ];
	}

	private final void drawBGBackCharFlipRotate() {
		int pat;
		final int buffer[] = this.buffer;
		final int transparentColor = paletteData[palette[16]];
		final int[] pal2rgb = this.pal2rgb;
		final int bgChrRom32Index = this.bgChrRom32Index;
		
		pat = chrRom32[bgChrRom32Index];
		if (buffer[63] == transparentColor) buffer[63] = pal2rgb[(pat       ) & 3];
		if (buffer[62] == transparentColor) buffer[62] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[61] == transparentColor) buffer[61] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[60] == transparentColor) buffer[60] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[59] == transparentColor) buffer[59] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[58] == transparentColor) buffer[58] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[57] == transparentColor) buffer[57] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[56] == transparentColor) buffer[56] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[55] == transparentColor) buffer[55] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[54] == transparentColor) buffer[54] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[53] == transparentColor) buffer[53] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[52] == transparentColor) buffer[52] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[51] == transparentColor) buffer[51] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[50] == transparentColor) buffer[50] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[49] == transparentColor) buffer[49] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[48] == transparentColor) buffer[48] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 1];
		if (buffer[47] == transparentColor) buffer[47] = pal2rgb[(pat       ) & 3];
		if (buffer[46] == transparentColor) buffer[46] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[45] == transparentColor) buffer[45] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[44] == transparentColor) buffer[44] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[43] == transparentColor) buffer[43] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[42] == transparentColor) buffer[42] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[41] == transparentColor) buffer[41] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[40] == transparentColor) buffer[40] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[39] == transparentColor) buffer[39] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[38] == transparentColor) buffer[38] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[37] == transparentColor) buffer[37] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[36] == transparentColor) buffer[36] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[35] == transparentColor) buffer[35] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[34] == transparentColor) buffer[34] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[33] == transparentColor) buffer[33] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[32] == transparentColor) buffer[32] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 2];
		if (buffer[31] == transparentColor) buffer[31] = pal2rgb[(pat       ) & 3];
		if (buffer[30] == transparentColor) buffer[30] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[29] == transparentColor) buffer[29] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[28] == transparentColor) buffer[28] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[27] == transparentColor) buffer[27] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[26] == transparentColor) buffer[26] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[25] == transparentColor) buffer[25] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[24] == transparentColor) buffer[24] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[23] == transparentColor) buffer[23] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[22] == transparentColor) buffer[22] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[21] == transparentColor) buffer[21] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[20] == transparentColor) buffer[20] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[19] == transparentColor) buffer[19] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[18] == transparentColor) buffer[18] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[17] == transparentColor) buffer[17] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[16] == transparentColor) buffer[16] = pal2rgb[(pat >>> 30)    ];
		pat = chrRom32[bgChrRom32Index + 3];
		if (buffer[15] == transparentColor) buffer[15] = pal2rgb[(pat       ) & 3];
		if (buffer[14] == transparentColor) buffer[14] = pal2rgb[(pat >>>  2) & 3];
		if (buffer[13] == transparentColor) buffer[13] = pal2rgb[(pat >>>  4) & 3];
		if (buffer[12] == transparentColor) buffer[12] = pal2rgb[(pat >>>  6) & 3];
		if (buffer[11] == transparentColor) buffer[11] = pal2rgb[(pat >>>  8) & 3];
		if (buffer[10] == transparentColor) buffer[10] = pal2rgb[(pat >>> 10) & 3];
		if (buffer[9 ] == transparentColor) buffer[9 ] = pal2rgb[(pat >>> 12) & 3];
		if (buffer[8 ] == transparentColor) buffer[8 ] = pal2rgb[(pat >>> 14) & 3];
		if (buffer[7 ] == transparentColor) buffer[7 ] = pal2rgb[(pat >>> 16) & 3];
		if (buffer[6 ] == transparentColor) buffer[6 ] = pal2rgb[(pat >>> 18) & 3];
		if (buffer[5 ] == transparentColor) buffer[5 ] = pal2rgb[(pat >>> 20) & 3];
		if (buffer[4 ] == transparentColor) buffer[4 ] = pal2rgb[(pat >>> 22) & 3];
		if (buffer[3 ] == transparentColor) buffer[3 ] = pal2rgb[(pat >>> 24) & 3];
		if (buffer[2 ] == transparentColor) buffer[2 ] = pal2rgb[(pat >>> 26) & 3];
		if (buffer[1 ] == transparentColor) buffer[1 ] = pal2rgb[(pat >>> 28) & 3];
		if (buffer[0 ] == transparentColor) buffer[0 ] = pal2rgb[(pat >>> 30)    ];
	}
	
	public void stateLoad(DataInputStream input) throws IOException {
		controlRegister1 = input.readUnsignedByte();
		controlRegister2 = input.readUnsignedByte();
		statusRegister = input.readUnsignedByte();
		spriteMemoryAddress = input.readInt();
		bgScrollOffsetH = input.readUnsignedByte();
		bgScrollOffsetV = input.readUnsignedByte();
		ppuMemoryAddress = input.readInt();
		bgScrollRatch = input.readBoolean();
		bgScrollRatch = input.readBoolean();
		ppuMemoryAddressRatch = input.readBoolean();
		for (int i = 0; i < vram.length; i++) {
			vram[i] = input.readUnsignedByte();
		}
		for (int i = 0; i < palette.length; i++) {
			palette[i] = input.readUnsignedByte();
		}
		for (int i = 0; i < spriteRam.length; i++) {
			spriteRam[i] = input.readUnsignedByte();
		}	
	}
	  
	public void stateSave(DataOutputStream output) throws IOException {
		output.writeByte(controlRegister1);
		output.writeByte(controlRegister2);
		output.writeByte(statusRegister);
		output.writeInt(spriteMemoryAddress);
		output.writeByte(bgScrollOffsetH);
		output.writeByte(bgScrollOffsetV);
		output.writeInt(ppuMemoryAddress);
		output.writeBoolean(bgScrollRatch);
		output.writeBoolean(ppuMemoryAddressRatch);
		output.writeBoolean(isPpuReadFirst);
		for (int i = 0; i < vram.length; i++) {
			output.writeByte(vram[i]);
		}
		for (int i = 0; i < palette.length; i++) {
			output.writeByte(palette[i]);
		}
		for (int i = 0; i < spriteRam.length; i++) {
			output.writeByte(spriteRam[i]);
		}			
	}
	
//	private final void DrawBG(Graphics g) {
//		
//		// wi}XN 1̏ꍇA8sNZ`悵Ȃ
//		int bgOrgX = 0;
//		if ((controlRegister2 & (1 << 1)) != 0) {
//			bgOrgX = 1;
//		}
//		
//		// Mirrormask
//		if (nesFile.isHorizontalMirror()) {
//			vramMirrorMask = 2;
//		} else {
//			vramMirrorMask = 1;
//		}
//				
//		// \̈256x224`悷
//		for (int y = 1; y < 29; y++) {
//			int scrollH = scrollHBuffer[y];
//			int scrollV = scrollVBuffer[y];				
//			int dx = scrollHBufferBack[y] - scrollH;
//			int dy = scrollVBufferBack[y] - scrollV;
//			
//			if (dx != 0 || dy != 0) {
//				g.copyArea(SCREEN_OFFSET_X, SCREEN_OFFSET_Y + (y << 3), 256, 8, dx, dy);
//			}
//			
//			for (int x = bgOrgX; x < 32; x++) {
//				getVirtualPPUPixels(g, x, y, scrollH, scrollV, ppuControlReg1Buffer[y]);
//			}
//		}
//		
//		System.arraycopy(scrollHBuffer, 0, scrollHBufferBack, 0, scrollHBuffer.length);
//		System.arraycopy(scrollVBuffer, 0, scrollVBufferBack, 0, scrollVBuffer.length);
//		System.arraycopy(ppuControlReg1Buffer, 0, ppuControlReg1BufferBack, 0, ppuControlReg1Buffer.length);
//	}
	
//	//	 PPUpixel擾
//	private final void getVirtualPPUPixels(Graphics g, int xx, int yy, int scrollH, int scrollV, int controlReg1)
//	{
//		int x = (xx << 3) + (scrollH & 0xF8);
//		int y = (yy << 3) + (scrollV & 0xF8);
//		
//		// x,y̒lANameTable#0`#3̂ǂɃANZX邩𔻒肷
//		int nameNo1 = ((x < 256) ? 0 : 1) + ((y < 240) ? 0 : 2);
//		
//		// nameNo擾(0`3)
//		int nameNo2 = controlReg1 & 3;
//		
//		// ۂɃANZXNameTableNovZ
//		int nameNo = nameNo1 ^ nameNo2;
//				
//		// Mirroring𔽉f
//		nameNo &= vramMirrorMask;
//		
//		// NameTableAddress OffsetvZ(1024{)
//		int nameOffset = nameNo * 0x400;
//		
//		// x0`255ɂāAchrXɂ
//		int chrX = x & 0xFF;
//		
//		// y0`239ɂāAchrYɂ
//		int chrY = (y < 240) ? y : (y - 240);
//		
//		// nameTbl32x30Ȃ̂ŁA8Ŋ
//		int nameX = chrX >> 3;	
//		int nameY = chrY >> 3;
//		
//		// NameTableindexvZinameTbl32x30Ȃ̂ŁA5bit shift32{j
//		int nameIndex = (nameY << 5) + nameX;
//		
//		// AttributeTableAddress OffsetvZ
//		int attOffset = nameOffset + 0x3C0;
//		
//		// AttTable8x7?Ȃ̂ŁA4Ŋ
//		int attX = nameX >> 2;
//		int attY = nameY >> 2;
//		
//		// AttributeTable̒lo
//		int att = vram[attOffset + (attY << 3) + attX];
//		
//		// att̕Kv2bito
//		int attShift = ((nameY & 2) << 1) + (nameX & 2);
//		att = (att >> attShift) & 3;
//		
//		if (!(xx < 2 || xx > 29)) {
//			// XV̗L𒲂ׂ
//			if (!isDirty[nameIndex] && !isDirtyPaletteSet[att]) {
//				return;
//			}
//		}
//		isDirty[nameIndex] = false;
//
//		// NameTablecharNoo
//		int charNo = vram[nameOffset + nameIndex];
//		
//		// Rg[WX^1Bit4Awip^[e[uAhXw擾
//		// 0:$0000A1:$1000
//		int patternAdr = (controlRegister1 & (1 << 4)) << 8;
//		
//		int drawX = (xx << 3) + SCREEN_OFFSET_X - (scrollH & 7);
//		int drawY = (yy << 3) + SCREEN_OFFSET_Y - (scrollV & 7);
//		
//		// ʉߐFx^̃LN^ꍇ
//		if (isTransparentOnly[(patternAdr >> 4) + charNo]) {
//			g.setColor(paletteData[palette[16]]);
//			g.fillRect(drawX, drawY, 8, 8);
//			return;
//		}
//				
//		// att̒lŁAparetteTable擾
//		pal2rgb = pal2rgbTbl[att];
//		
//		// chrRom32Adderss offsetvZ
//		bgChrRom32Index = (patternAdr + (charNo << 4)) >> 2;
//				
//		DrawOneCharFlipNone();
//		g.setPixels(drawX, drawY, 8, 8, buffer, 0);
//	}
}